"""
Diffraction calculation functions.


"""
from __future__ import absolute_import, print_function, division

from dtmm.conf import cached_function, BETAMAX, FDTYPE, CDTYPE
from dtmm.wave import betaphi
from dtmm.data import refind2eps
from dtmm.tmm import phase_mat,  alphaffi, alphaf,  alphaEEi, tr_mat, alphaE
from dtmm.linalg import dotmdm, dotmf
from dtmm.fft import fft2, ifft2


import numpy as np


DIFRACTION_PARAMETERS = ("distance", "mode")#, "refind")

@cached_function
def diffraction_alphaf(shape, ks, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.), betamax = BETAMAX, out = None):

    ks = np.asarray(ks)
    ks = abs(ks)
    beta, phi = betaphi(shape,ks)
    epsv = np.asarray(epsv, CDTYPE)
    epsa = np.asarray(epsa, FDTYPE)
    
    mask0 = (beta >= betamax)
    
    
    alpha, f= alphaf(beta,phi,epsv,epsa,out = out) 

    out = (alpha,f)
    
    f[mask0] = 0.
    alpha[mask0] = 0.
    

    return out



@cached_function
def diffraction_alphaffi(shape, ks, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.), betamax = BETAMAX, out = None):


    ks = np.asarray(ks)
    ks = abs(ks)
    beta, phi = betaphi(shape,ks)
    
    mask0 = (beta >= betamax)


    alpha, f, fi = alphaffi(beta,phi,epsv,epsa, out = out) 


    out = (alpha,f,fi)
    
#    try:
#        b1,betamax = betamax
#        a = (betamax-b1)/(betamax)
#        m = tukey(beta,a,betamax)
#        np.multiply(f,m[...,None,None],f)
#    except:
#        pass

    fi[mask0] = 0.
    f[mask0] = 0.
    alpha[mask0] = 0.
    #return mask, alpha,f,fi
    
    return out

@cached_function
def E_diffraction_alphaEEi(shape, ks, epsv = (1,1,1), 
                            epsa = (0.,0.,0.), mode = +1, betamax = BETAMAX, out = None):

    ks = np.asarray(ks)
    ks = abs(ks)
    beta, phi = betaphi(shape,ks)

    mask0 = (beta >= betamax)#betamax)
            
    alpha, j, ji = alphaEEi(beta,phi,epsv,epsa, mode = mode, out = out) 
    ji[mask0] = 0.
    j[mask0] = 0.
    alpha[mask0] = 0.
    out = (alpha,j,ji)
    return out


@cached_function
def E_diffraction_alphaE(shape, ks, epsv = (1,1,1), 
                            epsa = (0.,0.,0.), mode = +1, betamax = BETAMAX, out = None):

    ks = np.asarray(ks)
    ks = abs(ks)
    beta, phi = betaphi(shape,ks)

    mask0 = (beta >= betamax)#betamax)
            
    alpha, j = alphaE(beta,phi,epsv,epsa, mode = mode, out = out) 
    j[mask0] = 0.
    alpha[mask0] = 0.
    out = (alpha,j)
    return out

  
def phase_matrix(alpha, kd, mode = None, mask = None, out = None):
    kd = np.asarray(kd, dtype = FDTYPE)
    out = phase_mat(alpha,kd[...,None,None], out = out)  
    if mode == "t" or mode == +1:
        out[...,1::2] = 0.
    elif mode == "r" or mode == -1:
        out[...,::2] = 0.
    if mask is not None:
        out[mask] = 0.
    return out  

@cached_function
def field_diffraction_matrix(shape, ks,  d = 1., epsv = (1,1,1), epsa = (0,0,0.), mode = "b", betamax = BETAMAX, out = None):
    """Build field diffraction matrix. 
    """
    
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)
    alpha, f, fi = diffraction_alphaffi(shape, ks, epsv = epsv, epsa = epsa, betamax = betamax)
    kd =ks * d
    pmat = phase_matrix(alpha, kd , mode = mode)
    
    return dotmdm(f,pmat,fi,out = out) 

#@cached_function
def field_thick_cover_diffraction_matrix(shape, ks,  d = 1., epsv = (1,1,1), epsa = (0,0,0.), d_cover = 0, epsv_cover = (1.,1.,1.), epsa_cover = (0.,0.,0.), mode = "b", betamax = BETAMAX, out = None):
    """Build field diffraction matrix. 
    """
    
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)
    alpha, f, fi = diffraction_alphaffi(shape, ks, epsv = epsv, epsa = epsa, betamax = betamax)
    alpha0, f0 = diffraction_alphaf(shape, ks, epsv = epsv_cover ,epsa = epsa_cover, betamax = betamax)
    
    
    alphac = alpha0 - alpha / 1.5
    alphac = alphac - alphac[...,0,0,:][...,None,None,:]
    
    #offset = (alphac.mean(axis = (-2,-3)))[...,None,None,:]
    
    #alphac = alphac - offset
    
    
    
    kd = ks * d_cover
    
    pmatc = phase_matrix(alphac, kd , mode = mode)

    kd = ks * d
    
    pmat = phase_matrix(alpha, kd , mode = mode)
    
    pmat = pmat * pmatc
    
    return dotmdm(f,pmat,fi,out = out) 


@cached_function
def E_diffraction_matrix(shape, ks,  d = 1., epsv = (1,1,1), epsa = (0,0,0.), mode = +1, betamax = BETAMAX, out = None):
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)
    alpha, j, ji = E_diffraction_alphaEEi(shape, ks, epsv = epsv, epsa = epsa, mode = mode, betamax = betamax)
    kd =ks * d
    pmat = phase_matrix(alpha, kd)
    return dotmdm(j,pmat,ji,out = out) 

@cached_function
def E_cover_diffraction_matrix(shape, ks,  n = 1., d_cover = 0, n_cover = 1.5, mode = +1, betamax = BETAMAX, out = None):
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(refind2eps((n,)*3),CDTYPE)
    epsa = np.asarray((0.,0.,0.), dtype = FDTYPE)
    epsv_cover = np.asarray(refind2eps((n_cover,)*3),CDTYPE)
    epsa_cover = np.asarray((0.,0.,0.), dtype = FDTYPE)    
    alpha, j = E_diffraction_alphaE(shape, ks, epsv = epsv, epsa = epsa, mode = mode, betamax = betamax)
    alpha0, j0, j0i = E_diffraction_alphaEEi(shape, ks, epsv = epsv_cover, epsa = epsa_cover, mode = mode, betamax = betamax)

    alphac = alpha0 - alpha * n / n_cover
    alphac = alphac - alphac[...,0,0,:][...,None,None,:]     
    
    kd =ks * d_cover
    pmat = phase_matrix(alphac, kd)
    return dotmdm(j0,pmat,j0i,out = out) 



#@cached_function
#def jones_transmission_matrix(shape, ks, epsv_in = (1.,1.,1.), epsa_in = (0.,0.,0.),
#                            epsv_out = (1.,1.,1.), epsa_out = (0.,0.,0.), mode = +1, betamax = BETAMAX, out = None):
#    
#    
#    alpha, fin,fini = diffraction_alphaffi(shape, ks, epsv = epsv_in, 
#                            epsa = epsa_in, betamax = betamax)
#    
#    alpha, fout,fouti = diffraction_alphaffi(shape, ks, epsv = epsv_out, 
#                            epsa = epsa_out, betamax = betamax)
#    
#    return transmission_mat(fin, fout, fini = fini, mode = mode, out = out)
#
@cached_function
def E_tr_matrix(shape, ks, epsv_in = (1.,1.,1.), epsa_in = (0.,0.,0.),
                            epsv_out = (1.,1.,1.), epsa_out = (0.,0.,0.), mode = +1, betamax = BETAMAX, out = None):
    
    
    alpha, fin,fini = diffraction_alphaffi(shape, ks, epsv = epsv_in, 
                            epsa = epsa_in, betamax = betamax)
    
    alpha, fout,fouti = diffraction_alphaffi(shape, ks, epsv = epsv_out, 
                            epsa = epsa_out, betamax = betamax)
    
    return tr_mat(fin, fout, fmatini = fini, mode = mode, out = out)
#
#@cached_function
#def jones_t_matrix(shape, ks, epsv_in = (1.,1.,1.), epsa_in = (0.,0.,0.),
#                            epsv_out = (1.,1.,1.), epsa_out = (0.,0.,0.), mode = +1, betamax = BETAMAX, out = None):
#    
#    
#    alpha, fin,fini = diffraction_alphaffi(shape, ks, epsv = epsv_in, 
#                            epsa = epsa_in, betamax = betamax)
#    
#    alpha, fout,fouti = diffraction_alphaffi(shape, ks, epsv = epsv_out, 
#                            epsa = epsa_out, betamax = betamax)
#    
#    return t_mat(fin, fout, fini = fini, mode = mode, out = out)
        

@cached_function
def projection_matrix(shape, ks, epsv = (1,1,1),epsa = (0,0,0.), mode = +1, betamax = BETAMAX, out = None):
    """Computes a reciprocial field projection matrix.
    """
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)    
    alpha, f, fi = diffraction_alphaffi(shape, ks, epsv = epsv, epsa = epsa, betamax = betamax)
    kd = np.zeros_like(ks)
    pmat = phase_matrix(alpha, kd , mode = mode)
    return dotmdm(f,pmat,fi,out = out)   
 
  
def diffract(field, dmat, window = None, input_fft = False, output_fft = False, out = None):
    """Takes input field vector and diffracts it
    
    Parameters
    ----------
    field : (...,4,:,:) array
        Input field array.
    dmat : array
        Diffraction matrix. Use :func:`field_diffraction_matrix`to create one
    window : array, optional
        A window function applied to the result
    input_fft : bool
        Specifies whether input field array is the Fourier transform of the field.
        By default, input is assumed to be in real space.
    output_fft : bool
        Specifies whether the computed field array is the Fourier transform.
        By default, output is assumed to be in real space.
    out : ndarray, optional
        If specified, store the results here.
        
    Returns
    -------
    field : (...,4,:,:) array
        Diffracted field in real space (if output_fft == False )or in fft spaace
        (if output_fft == True).
            
    """
    if not input_fft:   
        field = fft2(field, out = out)
        out = field
    if dmat is not None:
        out = dotmf(dmat, field ,out = out)
    else:
        #make it work for dmat = None input... so just copy data
        if out is None:
            out = field.copy()
        else:
            if out is not field:
                out[...] = field
    if not output_fft:
        out = ifft2(out, out = out)
    if window is not None:
        if output_fft:
            raise ValueError("Cannot use window function if ouput is fft field.")
        out = np.multiply(out,window,out = out)
    return out

def diffracted_field(field, wavenumbers, d = 0.,n = 1, mode = "t", betamax = BETAMAX, out = None):
    eps = refind2eps([n]*3)
    pmat = field_diffraction_matrix(field.shape[-2:], wavenumbers, d = d, epsv = eps, epsa = (0.,0.,0.), mode = mode, betamax = betamax)
    return diffract(field, pmat, out = out) 
#
#def transmitted_field(field, wavenumbers, n = 1, betamax = BETAMAX, out = None):
#    eps = refind2eps([n]*3)
#    pmat = projection_matrix(field.shape[-2:], wavenumbers, epsv = eps, epsa = (0.,0.,0.), mode = "t", betamax = betamax)
#    return diffract(field, pmat, out = out) 
#
#def reflected_field(field, wavenumbers, n = 1, betamax = BETAMAX, out = None):
#    eps = refind2eps([n]*3)
#    pmat = projection_matrix(field.shape[-2:], wavenumbers, epsv = eps, epsa = (0.,0.,0.), mode = "r", betamax = betamax)
#    return diffract(field, pmat, out = out) 

__all__ = []



import unittest
import numpy as np
import dtmm.linalg as linalg


def vector2diagonal_matrix(vector):
    """
    Transforms a vector to a diagonal matrix.
    Parameters
    ----------
    vector : array
        Vector to turn into a diagonal matrix.

    Returns
    -------
    diagonal_matrix: array
        The diagonal matrix created from vector
    """
    # Create the diagonal matrix and set all values to 0
    diagonal_matrix = np.zeros(vector.shape + (vector.shape[-1],), vector.dtype)
    # Add vector elements along the diagonal
    for i in range(4):
        diagonal_matrix[..., i, i] = vector[..., i]

    return diagonal_matrix


class TestLinalg2(unittest.TestCase):
    
    def setUp(self):
        self.ni = 15
        self.nj = 16
        self.f = np.random.randn(4,self.ni,self.nj)+0j
        self.a = np.random.randn(self.ni,self.nj,4,4)+0j
        self.d = np.random.randn(self.ni,self.nj,4)+0j
        self.b = np.random.randn(self.ni,self.nj,4,4)+0j

    def dotmf(self,f,i,j,*matrices):
        mij = [m[i,j] for m in matrices] + [f[:,i,j]]
        return np.linalg.multi_dot(mij)
    
    def dot_multi(self,i,j,matrices):
        ms = [m[i,j].copy() for m in matrices[0:-1]] + [matrices[-1][:,i,j]]
        return np.linalg.multi_dot(ms)        
    
    def compare_results(self,out, matrices):
        for i in range(self.ni):
            for j in range(self.nj):   
                out2 = self.dot_multi(i,j,matrices)
                self.assertTrue(np.allclose(out[:,i,j],out2))        
        
    def test_dotmf(self):
        out = linalg.dotmf(self.a,self.f)
        self.compare_results(out,[self.a, self.f])
  
    def test_dotmdmf(self):
        out = linalg.dotmdmf(self.a,self.d,self.b,self.f)
        matrices = [self.a, vector2diagonal_matrix(self.d), self.b, self.f]
        self.compare_results(out,matrices)
        
    def test_dotmdmf2(self):
        kd = 2.3
        e = np.exp(1j*kd*self.d)
        out = linalg.dotmdmf(self.a,e,self.b,self.f)
        matrices = [self.a, vector2diagonal_matrix(e), self.b, self.f]
        self.compare_results(out,matrices)        

#    def test_ftransmit(self):
#        kd = 2.3
#        out = linalg.ftransmit(kd,self.a,self.d.real,self.b,self.f)
#        matrices = [self.a, vector2diagonal_matrix(np.exp(1j * kd * self.d.real)), self.b, self.f]
#        self.compare_results(out,matrices)

#    def test_btransmit(self):
#        kd = 2.3
#        out = linalg.btransmit(kd,self.a,self.d,self.b,self.f)
#        out2 = linalg.ftransmit(-kd,self.a,self.d,self.b,self.f)
#        self.assertTrue(np.allclose(out,out2)) 

                
if __name__ == "__main__":
    unittest.main()



"""
Custom 2D FFT functions.

numpy, scipy and mkl_fft do not have fft implemented such that output argument 
can be provided. This implementation adds the output argument for fft2 and 
ifft2 functions. 

Also, for mkl_fft and scipy, the computation can be performed in parallel using ThreadPool.

"""
from __future__ import absolute_import, print_function, division

from dtmm.conf import DTMMConfig, CDTYPE, MKL_FFT_INSTALLED, SCIPY_INSTALLED
import numpy as np

import numpy.fft as npfft

from multiprocessing.pool import ThreadPool

from functools import reduce

if MKL_FFT_INSTALLED == True:
    import mkl_fft
    
if SCIPY_INSTALLED == True:
    import scipy.fftpack as spfft

def _set_out_mkl(a,out):
    if out is not a:
        if out is None:
            out = a.copy()
        else:
            out[...] = a 
    return out

def _set_out(a,out):
    if out is not a:
        if out is None:
            out = np.empty_like(a)
    return out

def _copy(x,y):
    y[...] = x
    return y

_copy_if_needed = lambda x,y: _copy(x,y) if x is not y else x
    
def _sequential_inplace_fft(fft,array):
    [fft(d,overwrite_x = True) for d in array] 

def _sequential_fft(fft,array, out, overwrite_x = False):
    if out is array:
        [_copy_if_needed(fft(d, overwrite_x = True),out[i]) for i,d in enumerate(array)] 
    else:
        [_copy_if_needed(fft(d, overwrite_x = overwrite_x),out[i]) for i,d in enumerate(array)] 

def _optimal_workers(size, nthreads):
    if size%nthreads == 0:
        return nthreads
    else:
        return _optimal_workers(size, nthreads-1)
    
def _reshape(a):
    shape = a.shape
    newshape = reduce((lambda x,y: x*y), shape[:-2] or [1])
    if DTMMConfig.nthreads > 1:
        n = _optimal_workers(newshape,DTMMConfig.nthreads)
        newshape = (n,newshape//n,) + shape[-2:]
    else:
        newshape = (newshape,) + shape[-2:]
    a = a.reshape(newshape)
    return shape, a    

def __mkl_fft(fft,a,out):
    out = _set_out_mkl(a,out)
    shape, out = _reshape(out)
    if DTMMConfig.nthreads > 1:
        pool = ThreadPool(DTMMConfig.nthreads)
        workers = [pool.apply_async(_sequential_inplace_fft, args = (fft,d)) for d in out] 
        results = [w.get() for w in workers]
        pool.close()
    else:
        _sequential_inplace_fft(fft,out)
    return out.reshape(shape)    

def _mkl_fft2(a,out = None):
    return __mkl_fft(mkl_fft.fft2,a,out)

def _mkl_ifft2(a,out = None):
    return __mkl_fft(mkl_fft.ifft2,a,out)


def __sp_fft(fft,a,out, overwrite_x = False):
    out = _set_out(a,out)
    shape, a = _reshape(a)
    shape, out = _reshape(out)
    if DTMMConfig.nthreads > 1:
        pool = ThreadPool(DTMMConfig.nthreads)
        workers = [pool.apply_async(_sequential_fft, args = (fft,d,out[i],overwrite_x)) for i,d in enumerate(a)] 
        results = [w.get() for w in workers]
        pool.close()
    else:
        _sequential_fft(fft,a,out,overwrite_x)
    return out.reshape(shape)   

#def __sp_fft(fft,a,out):
#    if out is None:
#        return fft(a)
#    elif out is a:
#        out = fft(a, overwrite_x = True)
#        if out is a:
#            return out
#        else:
#            a[...] = out
#        return a
#    else:
#        out[...] = fft(a)
#        return out
        
def _sp_fft2(a, out = None):
    return __sp_fft(spfft.fft2, a, out)
    
def _sp_ifft2(a, out = None):
    return __sp_fft(spfft.ifft2, a, out)        

def __np_fft(fft,a,out):
    if out is None:
        return fft(a)
    else:
        out[...] = fft(a)
        return out
        
def _np_fft2(a, out = None):
    return __np_fft(npfft.fft2, a, out)
    
def _np_ifft2(a, out = None):
    return __np_fft(npfft.ifft2, a, out)       

                
def fft2(a, out = None):
    """Computes fft2 of the input complex array.
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    out : array or None, optional
        Output array. Can be same as input for fast inplace transform.
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the last two axes.
    """
    a = np.asarray(a, dtype = CDTYPE)
    libname = DTMMConfig["fftlib"]
    if libname == "mkl_fft":
        return _mkl_fft2(a, out)
    elif libname == "scipy":
        return _sp_fft2(a, out)
    elif libname == "numpy":
        return _np_fft2(a, out) 
    else:#default implementation is numpy
        return _np_fft2(a, out) 

    
def ifft2(a, out = None): 
    """Computes ifft2 of the input complex array.
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    out : array or None, optional
        Output array. Can be same as input for fast inplace transform.
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the last two axes.
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]
    if libname == "mkl_fft":
        return _mkl_ifft2(a, out)
    
    elif libname == "scipy":
        return _sp_ifft2(a, out)
    elif libname == "numpy":
        return _np_ifft2(a, out)  
    else: #default implementation is numpy
        return _np_ifft2(a, out)   

  
def mfft2(a, overwrite_x = False):
    """Computes matrix fft2 on a matrix of shape (..., n,n,4,4).
    
    This is identical to np.fft2(a, axes = (-4,-3))
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    overwrite_x : bool
        Specifies whether original array can be destroyed (for inplace transform)
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the (-4,-3) axes.    
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]    
    if libname == "mkl_fft":
        return mkl_fft.fft2(a, axes = (-4,-3), overwrite_x = overwrite_x)
    elif libname == "scipy":
        return spfft.fft2(a, axes = (-4,-3), overwrite_x = overwrite_x)
    elif libname == "numpy":
        return npfft.fft2(a, axes = (-4,-3))
    else: #default implementation is numpy
        return npfft.fft2(a, axes = (-4,-3))

def mifft2(a, overwrite_x = False):
    """Computes matrix ifft2 on a matrix of shape (..., n,n,4,4).
    
    This is identical to np.ifft2(a, axes = (-4,-3))
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    overwrite_x : bool
        Specifies whether original array can be destroyed (for inplace transform)
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the (-4,-3) axes.    
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]    
    if libname == "mkl_fft":
        return mkl_fft.ifft2(a, axes = (-4,-3), overwrite_x = overwrite_x)
    elif libname == "scipy":
        return spfft.ifft2(a, axes = (-4,-3), overwrite_x = overwrite_x)
    elif libname == "numpy":
        return npfft.ifft2(a, axes = (-4,-3))
    else: #default implementation is numpy
        return npfft.ifft2(a, axes = (-4,-3))    
    
def mfft(a, overwrite_x = False):
    """Computes matrix fft on a matrix of shape (..., n,4,4).
    
    This is identical to np.fft2(a, axis = -3)
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    overwrite_x : bool
        Specifies whether original array can be destroyed (for inplace transform)
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the (-4,-3) axes.    
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]    
    if libname == "mkl_fft":
        return mkl_fft.fft(a, axis = -3, overwrite_x = overwrite_x)
    elif libname == "scipy":
        return spfft.fft(a, axis = -3, overwrite_x = overwrite_x)
    elif libname == "numpy":
        return npfft.fft(a, axis = -3)
    else: #default implementation is numpy
        return npfft.fft(a, axis = -3)
    
def fft(a, overwrite_x = False):
    """Computes  fft on a matrix of shape (..., n).
    
    This is identical to np.fft2(a)
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    overwrite_x : bool
        Specifies whether original array can be destroyed (for inplace transform)
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the (-4,-3) axes.    
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]    
    if libname == "mkl_fft":
        return mkl_fft.fft(a, overwrite_x = overwrite_x)
    elif libname == "scipy":
        return spfft.fft(a,  overwrite_x = overwrite_x)
    elif libname == "numpy":
        return npfft.fft(a)
    else: #default implementation is numpy
        return npfft.fft(a)
    
def ifft(a, overwrite_x = False):
    """Computes  ifft on a matrix of shape (..., n).
    
    This is identical to np.ifft2(a)
    
    Parameters
    ----------
    a : array_like
        Input array (must be complex).
    overwrite_x : bool
        Specifies whether original array can be destroyed (for inplace transform)
       
    Returns
    -------
    out : complex ndarray
        Result of the transformation along the (-4,-3) axes.    
    """
    a = np.asarray(a, dtype = CDTYPE)      
    libname = DTMMConfig["fftlib"]    
    if libname == "mkl_fft":
        return mkl_fft.ifft(a, overwrite_x = overwrite_x)
    elif libname == "scipy":
        return spfft.ifft(a,  overwrite_x = overwrite_x)
    elif libname == "numpy":
        return npfft.ifft(a)
    else: #default implementation is numpy
        return npfft.ifft(a)


    
from setuptools import setup, find_packages
from dtmm import __version__

long_description = """
dtmm is an electro-magnetic field transmission and reflection calculation engine and visualizer. It can be used for calculation of transmission or reflection properties of layered homogeneous or inhomogeneous materials, such as confined liquid-crystals with homogeneous or inhomogeneous director profile. DTMM stands for Diffractive Transfer Matrix Method and is an adapted Berreman 4x4 transfer matrix method and an adapted 2x2 extended Jones method.
"""

packages = find_packages()

setup(name = 'dtmm',
      version = __version__,
      description = 'Diffractive Transfer Matrix Method',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author = 'Andrej Petelin',
      author_email = 'andrej.petelin@gmail.com',
      url="https://github.com/IJSComplexMatter/dtmm",
      packages = packages,
      #include_package_data=True
      package_data={
        # If any package contains *.dat, or *.ini include them:
        '': ['*.dat',"*.ini"]},
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
      )


"""
Field viewer
============

Matplotlib-based field visualizer (polarizing miscroscope simulator) and pom
image calculation functions

High level functions
--------------------

* :func:`.pom_viewer` for polarizing optical microscope simulation.
* :func:`.field_viewer` for raw field_data visualization.
* :func:`.bulk_viewer` for raw bulk_data visualization. 
* :func:`.calculate_pom_field` calculates polarizing optical microscope field.

Classes
-------

* :class:`.FieldViewer` is the actual field viewer object.
* :class:`.BulkViewer` is the actual bulk viewer object.
* :class:`.POMViewer` is the actual microscope viewer object.

"""

from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, AxesWidget, RadioButtons
from matplotlib.image import imsave
import scipy.ndimage as nd


#from dtmm.project import projection_matrix, project
from dtmm.color import load_tcmf, specter2color
from dtmm.diffract import diffract, field_diffraction_matrix, E_cover_diffraction_matrix, E_diffraction_matrix, E_tr_matrix
from dtmm.jones4 import ray_jonesmat4x4, mode_jonesmat4x4, mode_jonesmat2x2, ray_jonesmat2x2
from dtmm.field import field2specter, field2jones, jones2field
from dtmm.wave import k0
from dtmm.data import refind2eps
from dtmm.conf import BETAMAX, CDTYPE, get_default_config_option, DTMMConfig
from dtmm.jones import jonesvec
from dtmm import jones

from dtmm.linalg import dotmf, dotmm, dotmv
from dtmm.fft import fft2, ifft2

#: settable viewer parameters
VIEWER_PARAMETERS = ("focus","analyzer", "polarizer", "sample", "intensity","aperture", "retarder")
IMAGE_PARAMETERS = ("cols","rows","gamma","gray","cms")


SAMPLE_LABELS = ("-90 ","-45 "," 0  ","+45 ", "+90 ")
SAMPLE_NAMES = tuple((s.strip() for s in SAMPLE_LABELS))
POLARIZER_LABELS = (" H  "," V  ","LCP ","RCP ","none")
POLARIZER_NAMES = tuple((s.strip().lower() for s in POLARIZER_LABELS))
RETARDER_LABELS = ("$\lambda/4$", "$\lambda/2$","none")
RETARDER_NAMES = ("lambda/4", "lambda/2", "none")

def calculate_pom_field(field, jvec = None, pmat = None, dmat = None, window = None, input_fft = False, output_fft = False, out = None):
    """Calculates polarizing optical microscope field from the input field.
    
    This function refocuses the field, applies polarizer and analayzers
    
    Parameters
    ----------
    field : ndarray
        Input array of shape (...,:,4,:,:) describing multiwavelength polarized 
        field array or an array of shape (...,2,:,4,:,:) describing unpolarized 
        (x nad y polarized) multiwavelength field arrays. Works also with jones 
        multiwavelengths fields of shapes (...,:,2,:,:) and (...,2,:,2,:,:).
    jvec : jonesvec, optional
        Normalized jones vector describing which polarization state of the input
        field to choose. Input field must be of unpolarized type if this parameter
        is specified. 
    pmat : ndarray, optional
       A 4x4 or 2x2 matrix describing the analyzer and retarder matrix. This matrix
       is applied in real space only if both input_fft and output_fft are False,
       otherwise, the matrix is applied in Fourier space.
    dmat : ndarray, optional
       A diffraction matrix.
    window : array, optional
        If specified, windowing is applied after field is diffracted.
    input_fft : bool
        If specified, it idicates that we are working with fft data. pmat must
        be computed with mode_jonesmat4x4.
    output_fft : bool
        If specified, output data is left in FFT space. No inverse Fourier transform
        is performed if this parameter is set to True.
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    pom_field : ndarray
        Computed field od
        
    Examples
    --------
    >>> polarizer_jvec = dtmm.jones4.jonesvec((1,0))
    >>> analyzer_jvec = dtmm.jones4.jonesvec((0,1))
    >>> fmat = dtmm.tmm.f_iso()
    >>> pmat = dtmm.jones4.polarizer4x4(analyzer_jvec,fmat)
    >>> field_out = calculate_pom_field(field_in, polarizer_jvec, pmat)
    
    """
    if jvec is not None:
        if field.shape[-5] != 2:
            raise ValueError("Invalid field shape.")
        c,s = jvec
        x = field[...,0,:,:,:,:]*c
        y = np.multiply(field[...,1,:,:,:,:], s, out = out)
        field = np.add(x,y, out = out)#numexpr.evaluate("x*c+y*s", out = out)
    
    if input_fft == True or output_fft == True:
        #we are working on fft data, applying polarizer matrix in fft space
    
        if pmat is not None and dmat is not None:
            tmat = dotmm(pmat,dmat)
        elif pmat is None and dmat is not None:
            tmat = dmat
        elif pmat is not None and dmat is None and input_fft == True:
            tmat = pmat
        else:
            tmat = np.asarray(np.diag((1,1,1,1)), CDTYPE) 
            
        diffract(field,tmat,window = window, input_fft = input_fft, output_fft = output_fft, out = out)
    else:
        diffract(field,dmat,window = window, input_fft = input_fft, output_fft = output_fft,  out = out)
        #apply polarizer matrix in real space
        if pmat is not None:
            dotmf(pmat, field, out = out)
    return out

class CustomRadioButtons(RadioButtons):

    def __init__(self, ax, labels, active=0, activecolor='blue', size=49,
                 orientation="horizontal", **kwargs):
        """
        Add radio buttons to an `~.axes.Axes`.
        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            The axes to add the buttons to.
        labels : list of str
            The button labels.
        active : int
            The index of the initially selected button.
        activecolor : color
            The color of the selected button.
        size : float
            Size of the radio buttons
        orientation : str
            The orientation of the buttons: 'vertical' (default), or 'horizontal'.
        Further parameters are passed on to `Legend`.
        """
        AxesWidget.__init__(self, ax)
        self.activecolor = activecolor
        axcolor = ax.get_facecolor()
        self.value_selected = None

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_navigate(False)

        circles = []
        for i, label in enumerate(labels):
            if i == active:
                self.value_selected = label
                facecolor = activecolor
            else:
                facecolor = axcolor
            p = ax.scatter([],[], s=size, marker="o", edgecolor='black',
                           facecolor=facecolor)
            circles.append(p)
        if orientation == "horizontal":
            kwargs.update(ncol=len(labels), mode="expand")
        kwargs.setdefault("frameon", False)    
        self.box = ax.legend(circles, labels, loc="center", **kwargs)
        self.labels = self.box.texts
        self.circles = self.box.legendHandles
        for c in self.circles:
            c.set_picker(5)
        self.cnt = 0
        self.observers = {}

        self.connect_event('pick_event', self._clicked)


    def _clicked(self, event):
        if (self.ignore(event) or event.mouseevent.button != 1 or
            event.mouseevent.inaxes != self.ax):
            return
        if event.artist in self.circles:
            self.set_active(self.circles.index(event.artist))

def _redim(a, ndim=1):
    """Reshapes dimensions of input array by flattenig over first few dimensions. If
    ndim is larger than input array ndim, it adds new axes to input array.
    
    >>> a = np.zeros((4,5,6,7))
    >>> _redim(a,ndim = 3).shape
    (20,6,7)
    >>> _redim(a,ndim = 5).shape
    (1,4,5,6,7)
    
    """
    n = a.ndim - ndim 
    old_shape = a.shape
    if n < 0:
        new_shape = (1,)*abs(n) + old_shape
    else:
        new_shape = (np.multiply.reduce(old_shape[0:n+1]),) + old_shape[n+1:]
    return a.reshape(new_shape)


def bulk_viewer(field_data, **kwargs):
    """
    Returns a BulkViewer object for bulk field data visualization. See 
    :func:`.field_viewer` for parameters.
    
    Returns
    -------
    out : BulkViewer
        A :class:`BulkViewer` viewer object 
    
    """    
    return field_viewer(field_data, bulk_data=True, **kwargs)

def field_viewer(field_data, cmf=None, bulk_data=False, n=1., mode=None, is_polarized = None, 
                 window=None, diffraction=True, polarization_mode="normal", betamax=BETAMAX, beta = None, **parameters):
    """
    Returns a FieldViewer object for field data visualization.
    
    Parameters
    ----------
    field_data : tuple[np.ndarray]
        Input field data
    cmf : str, ndarray or None, optional
        Color matching function (table). If provided as an array, it must match 
        input field wavelengths. If provided as a string, it must match one of 
        available CMF names or be a valid path to tabulated data. See load_tcmf.
    bulk_data: bool
        Specifies whether data is to be treated as bulk data, e.g as returned by the
        :func:`.transfer.transfer_field` function with `ret_bulk = True`.
    n : float, optional
        Refractive index of the output material. Set this to the value used in
        the calculation of the field.
    mode : [ 't' | 'r' | None], optional
        Viewer mode 't' for transmission mode, 'r' for reflection mode None for
        as is data (no projection calculation - default).
    is_polarized : bool, optional
        If specified, it defines whether the field is polarize or not. For 
        non-polarized fields, the field must be of shape [...,2,:,4,:,:]. If 
        not provided, the polarization state is guessed from the shape of the
        input data. Setting this to False(and having non-polarized field)
        will allow setting the polarizer and sample rotation.
    window : ndarray, optional
        Window function by which the calculated field is multiplied. This can 
        be used for removing artefact from the boundaries.
    diffraction : bool, optional
        Specifies whether field is treated as diffractive field or not (if it
        was calculated by diffraction > 0 algorithm or not). If set to False
        refocusing is disabled.
    polarization_mode : str, optional
        Defines polarization mode. That is, how the polarization of the light is
        treated after passing the analyzer. By default, polarizer is applied
        in real space (`normal`) which is good for normal (or mostly normal) 
        incidence light. You can use `mode` instead of `normal` for more 
        accurate, but slower computation. Here polarizers are applied to 
        mode coefficients in fft space. 
    betamax : float
        Betamax parameter used in the diffraction calculation function. With this
        you can simulate finite NA of the microscope (NA = betamax).
    parameters : kwargs, optional
        Extra parameters passed directly to the :meth:`FieldViewer.set_parameters`
        
    Returns
    -------
    out : FieldViewer
        A :class:`FieldViewer` or :class:`BulkViewer` viewer object 
    
    """
    # Valid polarization modes
    valid_polarization_modes = ("mode", "normal")
    # Extract components out of field_data
    field, wavelengths, pixelsize = field_data
    
    wavelengths = np.asarray(wavelengths)
    if wavelengths.ndim == 0:
        #make it 1D
        wavelengths = wavelengths[None]
    
    if not diffraction and mode is not None:
        import warnings
        warnings.warn("Diffraction has been enabled because projection mode is set!")
        diffraction = True

    # Check that the provided polarization mode is a value one
    if polarization_mode not in valid_polarization_modes:
        raise ValueError("Unknown polarization mode, should be one of {}".format(repr(valid_polarization_modes)))
    if is_polarized is None:
        #try to get polarization state from the input data
        is_polarized = not (field.ndim >= 5 and field.shape[-5] == 2) 
    # Ensure a color matching function will be used
    if cmf is None:
        cmf = load_tcmf(wavelengths)
    elif isinstance(cmf, str):
        cmf = load_tcmf(wavelengths, cmf=cmf)

    if not bulk_data:
        if field.ndim < 4:
            raise ValueError("Incompatible field shape")

        viewer = FieldViewer(field.shape[-2:], wavelengths, pixelsize, propagation_mode = mode,
                             diffraction=diffraction, is_polarized = is_polarized,
                             refractive_index = n,
                             polarization_mode=polarization_mode, betamax=betamax, beta = beta)

        viewer.field = field
        viewer.image_parameters.cmf = cmf
        viewer.image_parameters.window = window
        viewer.set_parameters(**parameters)
    else:
        if field.ndim < 5:
            raise ValueError("Incompatible field shape")

        parameters.setdefault("focus", 0)
        viewer = BulkViewer(field.shape[-2:], wavelengths, pixelsize, propagation_mode = mode,
                             diffraction=diffraction, is_polarized = is_polarized,
                             refractive_index = n,
                             polarization_mode=polarization_mode, betamax=betamax, beta = beta)
        viewer.field = field
        viewer.image_parameters.cmf = cmf
        viewer.image_parameters.window = window
        viewer.set_parameters(**parameters)
        
    if DTMMConfig.verbose > 1:
        print("Loading field viewer")
        print("------------------------------------")
        viewer.viewer_options.print_info()
        viewer.image_parameters.print_info()
        print("-------------------------------------")
            

    return viewer

def pom_viewer(field_data, cmf=None, n = None, immersion = None, n_cover = None, d_cover = None, mode = +1, 
                 
                 is_polarized = None, window=None, NA = None, beta = None, **parameters):
    """
    Returns a FieldViewer object for optical microscope simulation.
    
    Parameters
    ----------
    field_data : tuple[np.ndarray]
        Input field data tuple.
    cmf : str, ndarray or None, optional
        Color matching function (table). If provided as an array, it must match 
        input field wavelengths. If provided as a string, it must match one of 
        available CMF names or be a valid path to tabulated data. See load_tcmf.
    n : float, optional
        Refractive index of the output material. If not set, it is set to 1 
        if immerion == False or n_cover if immersion == True.
    n_cover : float
        Refractive index of the cover medium. To simulate thick cover you should prepare 
        simulation results with nout = n_cover, (or nin = n_cover, if in reflection mode), 
        set n = 1 and d_cover > 0.
    d_cover : float
        Thickness ot the thick isotropic layer (cover glass).When d_cover != 0, t
        his simulates thick isotropic layer effect.
    immersion : bool
        Specified whether immersion objective is being used or not. Note that
        setting the 'n' parameter defines the immersion refractive index in
        this case.
    NA : float
        Numerical aperture of the objective. 
    mode : [ 't' | 'r' | +1 | -1 ]
        Viewer mode 't' or +1 for transmission mode, 'r' or -1 for reflection mode.
    is_polarized : bool, optional
        If specified, it defines whether the field is polarize or not. For 
        non-polarized fields, the field must be of shape [...,2,:,4,:,:]. If 
        not provided, the polarization state is guessed from the shape of the
        input data. Setting this to False(and having non-polarized field)
        will allow setting the polarizer and sample rotation.
    window : ndarray, optional
        Window function by which the calculated field is multiplied. This can 
        be used for removing artefact from the boundaries.
    parameters : kwargs, optional
        Extra parameters passed directly to the :meth:`FieldViewer.set_parameters`
        
    Returns
    -------
    out : POMViewer
        A :class:`POMViewer` viewer object 
        
    """
    # Extract components out of field_data
    field, wavelengths, pixelsize = field_data
    
    wavelengths = np.asarray(wavelengths)
    if wavelengths.ndim == 0:
        #make it 1D
        wavelengths = wavelengths[None]


    # Ensure a color matching function will be used
    if cmf is None:
        cmf = load_tcmf(wavelengths)
    elif isinstance(cmf, str):
        cmf = load_tcmf(wavelengths, cmf=cmf)

    if is_polarized is None:
        #try to get polarization state from the input data
        is_polarized = not (field.ndim >= 5 and field.shape[-5] == 2) 
        
    
    parameters.setdefault("focus", 0)
    viewer = POMViewer(field.shape[-2:], wavelengths, pixelsize, propagation_mode = mode,
                         is_polarized = is_polarized,
                         refractive_index = n, n_cover = n_cover, d_cover = d_cover,
                         NA=NA, immersion = immersion, beta = beta)
    
    viewer.image_parameters.cmf = cmf
    viewer.image_parameters.window = window
    viewer.set_parameters(**parameters)

    
    if DTMMConfig.verbose > 1:
        print("Loading POM viewer")
        print("------------------------------------")
        viewer.viewer_options.print_info()
        print("------------------------------------")
            
    viewer.field = field


    return viewer

def _as_field_array(field, options):
    field = np.asarray(field, CDTYPE)
    shape = options.shape
    nk = len(options.wavenumbers)
    if field.ndim >= 4 and field.shape[-3] == 4 and field.shape[-2:] == shape and field.shape[-4] == nk:
        return field
    else:
        raise ValueError("Invalid field data shape.")
        
def _as_jones_array(field, options):
    field = np.asarray(field, CDTYPE)
    shape = options.shape
    nk = len(options.wavenumbers)
    if field.ndim >= 4 and field.shape[-3] == 2 and field.shape[-2:] == shape and field.shape[-4] == nk:
        return field
    else:
        raise ValueError("Invalid jones data shape.")
    
def _float_or_none(value):
    """
    Helper function to convert the passed value to a float and return it, or return None.
    """
    return float(value) if value is not None else None
 
def _jonesmatrix_type(value):
    if isinstance(value, str):
        name = value.lower().strip()
        if name in (POLARIZER_NAMES + RETARDER_NAMES + ("x","y")+RETARDER_LABELS):
            return name, _jmat_from_name(name)
        else:
            raise ValueError("Not a valid jones matrix")
    else:
        try:
            angle = _float_or_none(value)
            return angle, _jmat_from_angle(angle)
        except TypeError:
            jmat = np.asarray(value,CDTYPE)
            if jmat.shape != (2,2):
                raise ValueError("Not a valid jones matrix")
            return jmat, jmat
        
def _jonesvector_type(value):
    if isinstance(value, str):
        name = value.lower().strip()
        if name in (POLARIZER_NAMES + ("x","y")):
            return name, _jvec_from_name(name)
        else:
            raise ValueError("Unknown jones vector!")
    else:
        try:
            angle = _float_or_none(value)
            return angle, _jvec_from_angle(angle)
        except TypeError:
            jvec = np.asarray(value,CDTYPE)
            if jvec.shape != (2,):
                raise ValueError("Not a valid jones vector")
            return jvec, jvec

def _jmat_from_name(name):
    if name in ("qplate","$\lambda/4$"):
        return jones.quarter_waveplate(np.pi/4)
    if name in ("hplate","$\lambda/2$"):
        return jones.half_waveplate(np.pi/4)
    if name == "none":
        return None
    return jones.polarizer(_jvec_from_name(name)) 

def _jvec_from_name(name):
    if name in ("h","x"):
        return jones.jonesvec((1,0))
    if name in ("v","y"):
        return jones.jonesvec((0,1))
    if name == "rcp":
        return jones.jonesvec((1,-1j))
    if name == "lcp":
        return jones.jonesvec((1,1j))
    if name == "none":
        return None
          
def _jmat_from_angle(angle):
    return jones.polarizer(_jvec_from_angle(angle)) if angle is not None else None 

def _jvec_from_angle(angle):
    return jonesvec((np.cos(np.radians(angle)),np.sin(np.radians(angle)))) if angle is not None else None 

class BaseViewerOptions(object):
    #: whether input field is polarized or non-polarized
    is_polarized = True
    #: list of beta values for field aperture simulations (for multi-ray data input)
    beta = None
    #: field propagation medium refractive index. 
    refractive_index = 1.
    #: propagation mode (+1 or "t" for transmission or -1 or "r" for reflection mode)
    propagation_mode = +1
    #: pixel size
    pixel_size = None
    #: simulated wavelengths
    wavelengths = []
    
    @property
    def wavenumbers(self):
        """simulated wavenumbers"""
        return k0(self.wavelengths, self.pixel_size)

    @property
    def epsv(self):
        """epsilon eigenvalues of the field propagation medium"""
        return refind2eps((self.refractive_index,)*3)
 

class FieldViewerOptions(BaseViewerOptions):
    """These parameters are set at initialization of the FieldViewer object,
    and should not be changed afterwards"""
    #: polarization mode, either "normal" or "mode"
    polarization_mode = "normal"
    #: whether to simulate diffraction. For bulk data visualization this is set to False
    diffraction = True
    #: whether to try to preserve memory while doing calculations.
    preserve_memory = False
    #: viewer betamax value
    betamax = BETAMAX
    
    def print_info(self):
        print(" $ polarization mode: {}".format(self.polarization_mode))  
        print(" $ polarized input: {}".format(self.is_polarized))  
        print(" $ refractive index: {}".format(self.refractive_index))     
        print(" $ propagation mode: {}".format(self.propagation_mode))   
        print(" $ max beta: {}".format(self.betamax))         
        
 
class POMViewerOptions(BaseViewerOptions):
    """These parameters are set at initialization of the FieldViewer object,
    and should not be changed afterwards"""
    #: cover glass refractive index
    _n_cover = None
    #: cover glass thickness
    _d_cover = None
    #: whether we use immersion objective or not.
    _immersion = None
    #: NA of the objective
    _NA = None
    
    def print_info(self):
        print(" $ polarization mode: {}".format(self.polarization_mode))  
        print(" $ polarized input: {}".format(self.is_polarized))  
        print(" $ cover refractive index: {}".format(self.n_cover)) 
        print(" $ cover thickness: {} [mm]".format(self.d_cover)) 
        print(" $ immersion: {}".format(self.immersion))
        print(" $ medium refractive index: {}".format(self.refractive_index))
        print(" $ propagation mode: {}".format(self.propagation_mode))   
        print(" $ objective NA: {}".format(self.NA))    
    
    @property
    def preserve_memory(self):
        return False
    
    @property
    def betamax(self):
        return self.NA
    
    @property
    def diffraction(self):
        return True
    
    @property
    def polarization_mode(self):
        return "mode"
   
    @property
    def refractive_index(self):
        if self._refractive_index is None:
            return self.n_cover if self.immersion else 1.
        else:
            return self._refractive_index
        
    @refractive_index.setter
    def refractive_index(self, value):
        self._refractive_index = value
       
    @property
    def n_cover(self):
        return get_default_config_option("n_cover", self._n_cover)
        
    @n_cover.setter    
    def n_cover(self, value):
        self._n_cover = value   

    @property
    def d_cover(self):
        return get_default_config_option("d_cover", self._d_cover)
        
    @d_cover.setter    
    def d_cover(self, value):
        self._d_cover = value   

    @property
    def immersion(self):
        return get_default_config_option("immersion", self._immersion)
        
    @immersion.setter    
    def immersion(self, value):
        self._immersion = value   
        
    @property    
    def NA(self):
        return get_default_config_option("NA", self._NA)

    @NA.setter    
    def NA(self, value):
        self._NA = value       
        
 
class ImageParameters(object):
    """Image parameters are storred here"""
    _wavelengths = None
    #: whether to convert RGB to gray
    _gray = None
    #: gamma value
    _gamma = None
    
    _cmf = None
    #: number of rows for periodic structure multiplication
    cols = 1
    #: number of rows for periodic structure multiplication
    rows = 1
    #: window function applied to the calculated image
    window = None
    
    def print_info(self):
        cmf_name = self.cmf if isinstance(self.cmf, str) else "custom"
        print(" $ color matchin function: {}".format(cmf_name)) 
        print(" $ output gray: {}".format(self.gray)) 
        print(" $ gamma: {}".format(self.gamma))
    
    @property 
    def gray(self):
        return get_default_config_option("gray", self._gray)
    
    @gray.setter
    def gray(self, value):
        self._gray = value
    
    @property 
    def gamma(self):
        return get_default_config_option("gamma", self._gamma)
    
    @gamma.setter
    def gamma(self, value):
        self._gamma = value

    @property      
    def cmf(self):
        """Color matching function data"""
        out = self._cmf 
        #if it does not exist, create one.
        if out is None:
            out = load_tcmf(self._wavelengths) 
            self._cmf = out
        return out 
    
    @cmf.setter        
    def cmf(self, cmf):
        if len(cmf) != len(self._wavelengths):
            raise ValueError("Incompatible cmf!")
        self._cmf = cmf
    
class FieldViewer(object): 
    """Field viewer. See :func:`.field_viewer`"""  
    _field = None
    _ffield = None
    _retarder = "none"
    _retarder_jmat = None
    _focus = 0
    _polarizer = None
    _polarizer_jvec = None
    _sample = None
    _sample_angle = None
    _analyzer = "none"
    _analyzer_jmat = None
    _intensity = 1.
    _parameters = VIEWER_PARAMETERS
    _image_parameters = IMAGE_PARAMETERS
    _fmin = 0
    _fmax = 100
    _jvec = None
    _pmat = None
    _dmat = None
    _ofield = None
    _specter = None
    
        
    def __init__(self,shape, wavelengths, pixelsize, **kwargs):
        self.viewer_options = FieldViewerOptions()
        self.image_parameters = ImageParameters()
        
        self.viewer_options.wavelengths = wavelengths
        self.viewer_options.pixel_size = pixelsize 
        self.viewer_options.shape = shape 
        self.image_parameters._wavelengths = wavelengths
        
        for key, value in kwargs.items():
            setattr(self.viewer_options, key, value)
            
        self._aperture = None if self.viewer_options.beta is None else max(self.viewer_options.beta)
        
        if not self.viewer_options.is_polarized:
            self.polarizer = "none"
            self.sample = "none"  
            
    def print_info(self):
        print(" $ intensity: {}".format(self.intensity)) 
        print(" $ polarizer: {}".format(self.polarizer)) 
        print(" $ sample rotation: {}".format(self.sample)) 
        print(" $ retarder: {}".format(self.retarder)) 
        print(" $ analyzer: {}".format(self.analyzer)) 
        print(" $ focus: {}".format(self.focus))

    def _clear_all_field_data(self):
        self._field = None
        self._ffield = None  
        self._specter = None
        
    @property
    def field(self):
        if self._field is None:
            self._field = ifft2(self._ffield)
        return self._field

    @field.setter
    def field(self, value):
        field = _as_field_array(value, self.viewer_options)
        self._clear_all_field_data()
        self._field = field
        
    @property
    def ffield(self):
        """Fourier transform of the field"""
        if self._ffield is None:
            self._ffield = fft2(self._field)
        return self._ffield
    
    @ffield.setter
    def ffield(self, value):
        """Fourier transform of the field"""
        ffield = _as_field_array(value, self.viewer_options)
        self._clear_all_field_data()
        self._ffield = ffield
        
    @property
    def _default_fmin(self):
        return self.focus - 100
    
    @property
    def _default_fmax(self):
        return self.focus + 100    
    
    @property
    def focus(self):
        """Focus position, relative to the calculated field position."""
        return self._focus   
    
    @property
    def masked_field(self):
        if self.aperture is not None:
            mask = self.viewer_options.beta <= self.aperture
            return self.field[mask]
        else:
            return self.field
    
    @property
    def masked_ffield(self):
        """Fourier transform of the field"""
        if self.aperture is not None:
            mask = self.viewer_options.beta <= self.aperture
            return self.ffield[mask]
        else:
            return self.ffield

    @focus.setter     
    def focus(self, z):
        if self.viewer_options.diffraction == True or z is None:
            self._dmat = None
            self._specter = None
            self._focus = _float_or_none(z)
        else:
            raise ValueError("Cannot set focus of a non-diffractive field.")

    @property
    def sample(self):
        """Sample rotation angle"""
        return self._sample
    
    @property
    def sample_angle(self):
        """Sample rotation angle in degrees in float"""
        return self._sample_angle
    
    @sample.setter    
    def sample(self, angle):
        """Sample rotation angle in degrees, in float or as a string"""
        if angle is not None:
            if self.viewer_options.is_polarized:
                raise ValueError("Input field must be unpolarized to use sample parameter!")
            
        if isinstance(angle, str):
            labels = tuple((label.strip() for label in SAMPLE_LABELS))
            if angle.strip() not in labels:
                if angle in ("none","+0","-0"):
                    angle = "0"
                else:
                    raise ValueError("sample angle must be a float or any of {}".format(labels))
        self._pmat = None #force recalculation of pmat
        self._jvec = None
        self._specter = None
        
        self._sample_angle = _float_or_none(angle)
        self._sample = angle
           
    @property
    def polarizer_jvec(self):
        return self._polarizer_jvec
    
    @polarizer_jvec.setter 
    def polarizer_jvec(self, value):
        self.polarizer = value
        
    @property
    def analyzer_jmat(self):
        return self._analyzer_jmat
    
    @analyzer_jmat.setter 
    def analyzer_jmat(self, value):
        self.analyzer = value
    
    @property    
    def aperture(self):
        """Illumination field aperture"""
        return self._aperture
    
    @aperture.setter    
    def aperture(self, value):
        self._aperture = _float_or_none(value)
        self._specter = None

    @property
    def polarizer(self):
        """Polarizer angle. Can be 'h','v', 'lcp', 'rcp', 'none', angle float or a jones vector"""
        return self._polarizer
        
    @polarizer.setter 
    def polarizer(self, angle):
        if angle is not None:
            if self.viewer_options.is_polarized:
                raise ValueError("Input field must be unpolarized to use polarizer parameter!")

        self._polarizer, self._polarizer_jvec = _jonesvector_type(angle)
        self._jvec = None
        self._specter = None
        
    @property
    def analyzer(self):
        """Analyzer angle. Can be 'h','v', 'lcp', 'rcp', 'none', angle float or a jones vector"""
        return self._analyzer    
        
    @analyzer.setter   
    def analyzer(self, angle):
        self._analyzer, self._analyzer_jmat = _jonesmatrix_type(angle)
        self._pmat = None
        self._specter = None
 
    @property    
    def retarder_jmat(self):
        return self._retarder_jmat
    
    @property    
    def retarder(self):
        return self._retarder
    
    @retarder.setter    
    def retarder(self, value):
        self._retarder, self._retarder_jmat = _jonesmatrix_type(value)
        self._pmat = None #force recalculation of pmat
        self._specter = None
       
    @property
    def intensity(self):
        """Input light intensity"""
        return self._intensity   
    
    @intensity.setter   
    def intensity(self, intensity):
        self._intensity = _float_or_none(intensity)
    
    @property
    def input_jones(self):
        """Input field jones vector"""
        if self._jvec is None:
            sample = self.sample_angle
            if sample is None:
                sample = 0.
            if self.polarizer_jvec is None:
                self._jvec = None
            else:
                r = jones.rotation_matrix2(-np.radians(sample))
                self._jvec = dotmv(r,self.polarizer_jvec)
        return self._jvec
    
    @property
    def diffraction_matrix(self):
        """Diffraction matrix for diffraction calculation"""
        if self._dmat is None:
            vp = self.viewer_options
            if vp.diffraction or vp.propagation_mode is not None:
                #if mode is selected, we need to project the filed using diffraction
                d = 0 if self.focus is None else self.focus
                epsv = vp.epsv
                self._dmat = field_diffraction_matrix(vp.shape, vp.wavenumbers, d = d, 
                                          epsv = epsv, mode = vp.propagation_mode, betamax = vp.betamax) 
            else:
                self._dmat = None
        return self._dmat
    
    @property
    def output_matrix(self):
        """4x4 jones output matrix"""
        sample = self.sample_angle
        if sample is None:
            sample = 0.
        if self._pmat is None:
            if self.analyzer_jmat is not None:
                m = dotmm(self.analyzer_jmat,self.retarder_jmat) if self.retarder_jmat is not None else self.analyzer_jmat
                m = jones.rotated_matrix(m,np.radians(sample))       
                vp = self.viewer_options
                epsv = vp.epsv
                if vp.polarization_mode== "mode":                 
                    self._pmat = mode_jonesmat4x4(vp.shape, vp.wavenumbers, m, epsv = epsv)
                else:
                    self._pmat = ray_jonesmat4x4(m, epsv = epsv)
            else:
                self._pmat = None
        return self._pmat
        
    def set_parameters(self, **kwargs):
        """Sets viewer parameters. Any of the :attr:`.VIEWER_PARAMETERS`
        """
        for key, value in kwargs.items():
            if key in self._parameters:
                setattr(self, key, value) 
            elif key in self._image_parameters:
                setattr(self.image_parameters, key, value) 
            else:
                raise TypeError("Unexpected keyword argument '{}'".format(key))

    def get_parameters(self):
        """Returns viewer parameters as dict"""
        return {name : getattr(self,name) for name in VIEWER_PARAMETERS}, {name : getattr(self.image_parameters,name) for name in IMAGE_PARAMETERS}
        
    def plot(self, fig = None,ax = None, sliders = None, show_sliders = None, show_scalebar = None, show_ticks = None, **kwargs):
        """Plots field intensity profile. You can set any of the below listed
        arguments. Additionaly, you can set any argument that imshow of
        matplotlib uses (e.g. 'interpolation = "sinc"').
        
        Parameters
        ----------
        show_slider : bool, optional
            Specifies whether to show sliders or not.
        show_scalebar : bool, optional
            Specifies whether to show scalebar or not.
        show_ticks : bool, optional
            Specifies whether to show ticks in imshow or not.
        fmin : float, optional
            Minimimum value for the focus setting.
        fmax : float, optional
            Maximum value for the focus setting.             
        imin : float, optional
            Minimimum value for then intensity setting.
        imax : float, optional
            Maximum value for then intensity setting.     
        pmin : float, optional
            Minimimum value for the polarizer angle.
        pmax : float, optional
            Maximum value for the polarizer angle.    
        smin : float, optional
            Minimimum value for the sample rotation angle.
        smax : float, optional
            Maximum value for the sample rotation angle.  
        amin : float, optional
            Minimimum value for the analyzer angle.
        amax : float, optional
            Maximum value for the analyzer angle.  
        namin : float, optional
            Minimimum value for the numerical aperture.
        namax : float, optional
            Maximum value for the numerical aperture.  
        """
        show_sliders = get_default_config_option("show_sliders",show_sliders)
        show_ticks = get_default_config_option("show_ticks",show_ticks)
        show_scalebar = get_default_config_option("show_scalebar",show_scalebar)
        
        if fig is None:
            if ax is None:
                self.fig = plt.figure() 
            else:
                self.fig = ax.figure
        else:
            self.fig = fig
                
        self.ax = self.fig.add_subplot(111) if ax is None else ax
        
        plt.subplots_adjust(bottom=0.27)  
        image = self.calculate_image()
        
        self.sliders = {} if sliders is None else sliders
        
        if show_sliders:
            
            def update_slider_func(name):
                def update(d):
                    setattr(self,name,d)
                    self.update_plot()
                return update
                
            axes = [[0.25, 0.19, 0.65, 0.03],
                    [0.25, 0.16, 0.65, 0.03],
                    [0.25, 0.13, 0.65, 0.03],
                    [0.25, 0.10, 0.65, 0.03],
                    [0.25, 0.07, 0.65, 0.03],
                    [0.25, 0.04, 0.65, 0.03],
                    [0.25, 0.01, 0.65, 0.03],]
            
            def add_slider(name, axpos, names = None,labels = None, min_name = None, max_name = None, min_value = 0, max_value = 1, valfmt = '%.1f'):
                func = update_slider_func(name)
                if self.sliders.get(name) is None:
                    ax = self.fig.add_axes(axpos)
                    obj = getattr(self,name)
                    
                    if isinstance(obj, str):
                        if names is None:
                            names = tuple((label.strip().lower() for label in labels))
                        active = names.index(obj)
                        self.sliders[name] = CustomRadioButtons(ax, labels, active = active)
                        ax.set_ylabel(name,rotation="horizontal",ha = "right", va = "center")
                    else:
                        self.sliders[name] = Slider(ax, name,min(kwargs.pop(min_name,min_value),obj),max(kwargs.pop(max_name,max_value),obj),valinit = obj, valfmt=valfmt)
                try:
                    id = self.sliders[name].on_changed(func)
                except AttributeError:
                    id = self.sliders[name].on_clicked(func)
                    
                return id, ax

            if self.aperture is not None:   
                axpos = axes.pop() 
                self._ids6, self.axaperture = add_slider("aperture", axpos, labels = POLARIZER_LABELS, min_name = "namin", max_name = "namax", min_value = 0, max_value = self.viewer_options.beta.max())
                            
            if self.intensity is not None:
                axpos = axes.pop() 
                self._ids5, self.axintensity = add_slider("intensity", axpos, min_name = "imin", max_name = "imax", min_value = 0, max_value = self.intensity * 10)
 
            if self.polarizer is not None:
                axpos = axes.pop() 
                self._ids4, self.axpolarizer = add_slider("polarizer", axpos, labels = POLARIZER_LABELS, min_name = "pmin", max_name = "pmax", min_value = -90, max_value = 90)
 
            if self.sample is not None:
                axpos = axes.pop() 
                self._ids3, self.axsample = add_slider("sample", axpos,labels = SAMPLE_LABELS, min_name = "smin", max_name = "smax", min_value = -180, max_value = 180)
 
            if self.retarder is not None:
                axpos = axes.pop() 
                self._ids2, self.axretarder = add_slider("retarder", axpos, names = RETARDER_NAMES, labels = RETARDER_LABELS)
 
            if self.analyzer is not None:
                axpos = axes.pop() 
                self._ids1, self.axanalyzer = add_slider("analyzer", axpos, labels = POLARIZER_LABELS, min_name = "amin", max_name = "amax", min_value = -90, max_value = 90)
 
            if self.focus is not None:  
                axpos = axes.pop() 
                self._ids0, self.axfocus = add_slider("focus", axpos, min_name = "fmin", max_name = "fmax", min_value = self._default_fmin, max_value = self._default_fmax)
 
            
        self.axim = self.ax.imshow(image, origin = kwargs.pop("origin","lower"), **kwargs)
        
        if show_scalebar == True:
            if self.viewer_options.pixel_size is None:
                raise ValueError("You must provide pixel_size to show scale bar.")
            try:
                from matplotlib_scalebar.scalebar import ScaleBar
            except ImportError:
                raise ValueError("You must have matplotlib_scalebar installed to use this feature.")
            
            scalebar = ScaleBar(self.viewer_options.pixel_size , "nm")
            self.ax.add_artist(scalebar)
            
        show_ticks = False if show_scalebar and show_ticks is None else show_ticks 
        
        if show_ticks == False:
            self.ax.set_xticklabels([])
            self.ax.set_xticks([])
            self.ax.set_yticklabels([])
            self.ax.set_yticks([])            
        
        return self.ax.figure, self.ax
                    
    def _calculate_diffraction(self):  
        if self._dmat is None:
            self._ofield = None #we have to create new memory for output field  
            vp = self.viewer_options
            modal = vp.polarization_mode == "mode"
            diffractive = vp.diffraction == True
            preserve_memory = vp.preserve_memory
            if preserve_memory or modal or self.diffraction_matrix is None:
                # we calculate diffraction later or not at all
                self._ofield = self.masked_ffield if modal else self.masked_field 
            else:
                if diffractive or vp.propagation_mode is not None: 
                    self._ofield = diffract(self.masked_field,self.diffraction_matrix,window = self.image_parameters.window,out = self._ofield)
                else:
                    #no diffraction at all..
                    if self.image_parameters.window is not None:
                        self._ofield = self.masked_field * self.image_parameters.window
                    else:
                        self._ofield = self.masked_field.copy()
                    
    def _field2specter(self,field):
        return field2specter(field)
    
    def _calculate_pom_field(self, data, jvec, pmat, dmat, window = None, input_fft = False, out = None):
        return calculate_pom_field(data,jvec,pmat ,dmat,window = window,input_fft = input_fft, out = out)
        
    def _calculate_specter(self):
        if self._specter is None:
            vp = self.viewer_options
            modal = vp.polarization_mode == "mode"
            preserve_memory = vp.preserve_memory
            jvec = self.input_jones
            
            input_fft = True if modal else False 
            
            field = self._ofield
            
            if jvec is None:
                tmp = _redim(field, ndim = 5)
                out = np.empty_like(tmp[0])
          
            else:
                tmp = _redim(field, ndim = 6)
                out = np.empty_like(tmp[0,0])
    
            self._specter = 0.
            
            #we may heve calculated diffraction before, set to None if we had
            dmat = self.diffraction_matrix if preserve_memory or modal else None
            
            for i,data in enumerate(tmp): 
                field = self._calculate_pom_field(data,jvec,self.output_matrix,dmat,window = self.image_parameters.window,input_fft = input_fft, out = out)
                self._specter += self._field2specter(field)
                   
    def calculate_field(self, recalc = False, **params):
        self.set_parameters(**params)
        vp = self.viewer_options
        modal = vp.polarization_mode == "mode"
        preserve_memory = vp.preserve_memory

        self._calculate_diffraction()
            
        jvec = self.input_jones
        
        input_fft = True if modal else False 
        
        field = self._ofield
        
        #we may heve calculated diffraction before, set to None if we had
        dmat = self.diffraction_matrix if preserve_memory or modal else None

        return calculate_pom_field(field,jvec,self.output_matrix,dmat,window = self.image_parameters.window,input_fft = input_fft)
                    
    def calculate_specter(self, **params):
        """Calculates field specter.
        
        Parameters
        ----------
        params: kwargs, optional
            Any additional keyword arguments that are passed dirrectly to 
            set_parameters method.
        """        
        self.set_parameters(**params)
        if DTMMConfig.verbose > 1:
            print("Calculating specter.")
            print("------------------------------------")            
            self.print_info()
            print("------------------------------------")
                
        
        self._calculate_diffraction()
        self._calculate_specter()
        return self._specter
 
    def calculate_image(self, **params):
        """Calculates RGB image.
        
        Parameters
        ----------
        params: keyword arguments
            Any additional keyword arguments that are passed dirrectly to 
            set_parameters method.
            
        """   
        specter = self.calculate_specter(**params)
        
        if DTMMConfig.verbose > 1:
            print("Calculating image.")
            print("------------------------------------")
            self.image_parameters.print_info()
            print("------------------------------------")
        
        vp = self.viewer_options
        ip =  self.image_parameters
        
        cmf = ip.cmf
    
        if self.intensity is not None:
            if self.intensity != 0.0:
                if vp.propagation_mode in (-1,"r"):
                    #poynting is negative, make it positive
                    norm = -1./self.intensity
                else:
                    norm = 1./self.intensity
            else:
                norm = 0.0

            image = specter2color(specter,cmf, norm = norm, gray = ip.gray, gamma = ip.gamma) 
        else:
            if vp.propagation_mode in (-1,"r"):
                image = specter2color(specter,cmf, norm = -1., gray = ip.gray, gamma = ip.gamma) 
            else:
                image = specter2color(specter,cmf, gray = ip.gray, gamma = ip.gamma) 
        
        image = np.hstack(tuple((image for i in range (ip.cols))))
        image = np.vstack(tuple((image for i in range (ip.rows))))
        
        if self.sample_angle != 0 and self.sample_angle is not None:
            image = nd.rotate(image, -self.sample_angle, reshape = False, order = 1) 

        return image
 
    def save_image(self, fname, origin = "lower", **kwargs):
        """Calculates and saves image to file using matplotlib.image.imsave.
        
        Parameters
        ----------
        fname : str
            Output filename or file object.
        origin : [ 'upper' | 'lower' ]
            Indicates whether the (0, 0) index of the array is in the upper left 
            or lower left corner of the axes. Defaults to 'lower' 
        kwargs : optional
            Any extra keyword argument that is supported by matplotlib.image.imsave
        """
        im = self.calculate_image()
        imsave(fname, im, origin = origin, **kwargs)

    def update_plot(self):
        """Triggers plot redraw"""
        image = self.calculate_image()
        self.axim.set_data(image)
        #for key, slider in self.sliders.items():
        #    slider.set_active(False)
        #    slider.set_val(getattr(self,key))
        #    slider.set_active(True)
        self.fig.canvas.draw_idle()  
   
    def show(self):
        """Shows plot"""
        plt.show()

                 
class BulkViewer(FieldViewer):
    @property
    def _default_fmin(self):
        return 0
    
    @property
    def _default_fmax(self):
        return len(self.field) -1   
    
    @property
    def focus(self):
        """Focus position"""
        return self._focus       
    
    @focus.setter     
    def focus(self, z):
        #focos must be integer here, index of the layer
        i = int(z)
        #check is ok, raise IndexError else
        self.field[i]
        self._focus = i
        
    @property
    def masked_field(self):
        if self.aperture is not None:
            mask = self.viewer_options.beta <= self.aperture
            return self.field[self.focus,mask]
        else:
            return self.field[self.focus]
    
    @property
    def masked_ffield(self):
        """Fourier transform of the field"""
        if self.aperture is not None:
            mask = self.viewer_options.beta <= self.aperture
            return self.ffield[self.focus,mask]
        else:
            return self.ffield[self.focus]
        
class POMViewer(FieldViewer):
    """Similar to FieldViewer, with the following differences:
        
    Computation is done on jones field, instead of full field data. Therefore,
    you need to provide the mode parameter to select the propagation mode.
    
    """
    _fjones = None
    _jones = None
    _cmat = None
    
    def __init__(self,shape, wavelengths, pixelsize, **kwargs):
        self.viewer_options = POMViewerOptions()
        self.image_parameters = ImageParameters()
        
        self.viewer_options.wavelengths = wavelengths
        self.viewer_options.pixel_size = pixelsize 
        self.viewer_options.shape = shape 
        self.image_parameters._wavelengths = wavelengths
        
        for key, value in kwargs.items():
            setattr(self.viewer_options, key, value)
            
        self._aperture = None if self.viewer_options.beta is None else max(self.viewer_options.beta)
        
        if not self.viewer_options.is_polarized:
            self.polarizer = "none"
            self.sample = "none"  
        
    def _clear_all_field_data(self):
        self._jones = None
        self._fjones = None
        self._field = None
        self._ffield = None      
         
    @property
    def jones(self):
        if self._jones is None:   
            vp = self.viewer_options
            epsv = refind2eps((vp.n_cover,)*3)
            self._jones = field2jones(self._field, vp.wavenumbers, epsv = epsv, mode = vp.propagation_mode, output_fft = False, betamax = vp.betamax)
        return self._jones   
    
    @jones.setter
    def jones(self,value):
        self._clear_all_field_data()
        self._jones = value
        self._fjones = None
        self._field = None
        self._ffield = None
        
    
    @property
    def fjones(self):
        if self._fjones is None:
            vp = self.viewer_options
            epsv = refind2eps((vp.n_cover,)*3)
            if self._jones is None:
                self._fjones = field2jones(self._field, vp.wavenumbers, epsv = epsv, mode = vp.propagation_mode, output_fft = True, betamax = vp.betamax)
            else:
                self._fjones = fft2(self._jones)
        return self._fjones
    
    @fjones.setter
    def fjones(self, value):
        self._fjones = value
        self._field = None
        self._ffield = None
        self._jones = None

    @property
    def masked_fjones(self):
        """Fourier transform of the field"""
        if self.aperture is not None:
            mask = self.viewer_options.beta <= self.aperture
            return self.fjones[mask]
        else:
            return self.fjones
        
    def _calculate_diffraction(self):
        #diffraction is calculated during 
        self._ofield = self.masked_fjones

    def _field2specter(self,field):
        vp = self.viewer_options
        epsv = vp.epsv
        return field2specter(jones2field(field, vp.wavenumbers, epsv = epsv, mode = vp.propagation_mode, input_fft = True, betamax = vp.betamax))

    def _calculate_pom_field(self, data, jvec, pmat, dmat, window = None, input_fft = False, out = None):
        return calculate_pom_field(data,jvec,pmat ,dmat,window = window,input_fft = input_fft, output_fft = True, out = out)
      
    @property
    def output_matrix(self):
        """2x2 jones output matrix"""
        sample = self.sample_angle
        if sample is None:
            sample = 0.
        if self._pmat is None:
            vp = self.viewer_options
            if self.analyzer_jmat is not None:
                m = dotmm(self.analyzer_jmat,self.retarder_jmat) if self.retarder_jmat is not None else self.analyzer_jmat
                m = jones.rotated_matrix(m,np.radians(sample))
                epsv = vp.epsv
                self._pmat = mode_jonesmat2x2(vp.shape, vp.wavenumbers, m, mode = vp.propagation_mode, epsv = epsv, betamax = vp.betamax)

            else:
                self._pmat = None
        return self._pmat
    
    @property
    def cover_matrix(self):
        if self._cmat is None:
            vp = self.viewer_options
            # vp.d_cover is in mm, calculat d_cover in pixel units
            d_cover =vp.d_cover / vp.pixel_size * 1000000 
            cmat = E_cover_diffraction_matrix(vp.shape, vp.wavenumbers, 
                                      n = vp.refractive_index, d_cover = d_cover,n_cover = vp.n_cover,
                                      mode = vp.propagation_mode, betamax = vp.betamax) 
            # Fresnel reflection matrix
            eps = vp.epsv
            eps_cover = refind2eps((vp.n_cover,)*3)
            tmat,rmat = E_tr_matrix(vp.shape, vp.wavenumbers, epsv_in = eps_cover,
                        epsv_out = eps, mode = vp.propagation_mode, betamax = vp.betamax)
            #self._cmat = cmat
            self._cmat = dotmm(tmat,cmat)
            
        return self._cmat
    
    @property
    def diffraction_matrix(self):
        """Diffraction matrix for diffraction calculation"""
        if self._dmat is None:
            vp = self.viewer_options
            #if mode is selected, we need to project the filed using diffraction
            d = 0 if self.focus is None else self.focus
            epsv = vp.epsv
            #diffraction matrix
            dmat = E_diffraction_matrix(vp.shape, vp.wavenumbers,  d = d, 
                                      epsv = epsv, 
                                      mode = vp.propagation_mode, betamax = vp.betamax) 
            self._dmat = dmat

            cmat = self.cover_matrix  
            self._dmat = dotmm(self._dmat, cmat) 
        return self._dmat   

    
__all__ = ["calculate_pom_field", "field_viewer", "bulk_viewer", "FieldViewer", "BulkViewer"]
    


"""Tests for fft"""

import unittest
import numpy as np
from dtmm.tmm import f_iso, fvec, fvec2E, E2fvec
import dtmm.conf
from dtmm.conf import MKL_FFT_INSTALLED

class TestConversion(unittest.TestCase):
    
    def setUp(self):
        self.f0 = f_iso(beta = 0.2, phi = 0.2, n = 1.5)
        self.fvecp = fvec(self.f0, jones = (1,0), mode = +1)
        self.fvecm = fvec(self.f0, jones = (1,0), mode = -1)
        self.fvec = self.fvecp + self.fvecm
     
    def test_fvec2E(self):
        jp = fvec2E(self.fvec, self.f0, mode = +1)
        jm = fvec2E(self.fvec, self.f0, mode = -1)
        
        self.assertTrue(np.allclose(jp, self.fvecp[::2]))
        self.assertTrue(np.allclose(jm, self.fvecm[::2]))
        
        jp = fvec2E(self.fvecp, self.f0, mode = -1)
        jm = fvec2E(self.fvecm, self.f0, mode = +1)
        
        self.assertTrue(np.allclose(jp, np.asarray((0j,0j)),atol = 1e-6))
        self.assertTrue(np.allclose(jm, np.asarray((0j,0j)),atol = 1e-6)) 


                
if __name__ == "__main__":
    unittest.main()


"""
Field creation, transformation in IO functions
"""

from __future__ import absolute_import, print_function, division

import numpy as np

from dtmm.conf import NCDTYPE,NFDTYPE, FDTYPE, CDTYPE, NUMBA_PARALLEL, NUMBA_TARGET, NUMBA_CACHE, BETAMAX , DTMMConfig, get_default_config_option
from dtmm.wave import planewave, betaphi, wave2eigenwave
from dtmm.diffract import diffracted_field, diffraction_alphaf
from dtmm.window import aperture
from dtmm.fft import fft2, ifft2, fft, ifft
from dtmm.wave import betaxy, eigenmask, eigenmask1
from dtmm.window import blackman
from dtmm.tmm import alphaf ,fvec2E, E2fvec
from dtmm.tmm import fvec as field4
from dtmm.data import refind2eps
from dtmm.jones import jonesvec

import numba as nb
from numba import prange

if NUMBA_PARALLEL == False:
    prange = range
    
sqrt = np.sqrt

def field2fvec(field):
    """transposes field array from shape (..., k,n,m) to (...,n,m,k)."""
    return np.moveaxis(field,-3,-1)

def fvec2field(vec):
    """transposes vector from shape (..., n,m,k) to (...,k,n,m)"""
    return np.moveaxis(vec,-1,-3)

def field2jones(field, ks, beta = None, phi = None, epsv = (1.,1.,1.), epsa = (0.,0.,0.), mode = +1, input_fft = False, output_fft = False, betamax = BETAMAX):
    """Converts (..., 4,n,m) field array to (..., 2,n,m) jones array
    
    The conversion is done either in reciprocal space (default) or in real space,
    if you provide the beta or phi parameter.
    
    Parameters
    ----------
    field : ndarray
        Input field array data of shape (..., 4,n,m).
    ks : array
        A list of k-values (wave numbers).
    beta : float, optional
        If set, it defines the field beta parameter, conversion is then done
        in real space.
    phi : float, optional
        If set, it defines the field phi parameter, conversion is then done
        in real space.
    epsv : array
        The epsilon eigenvalues of the medium
    epsv : array
        The Euler rotation angles.
    mode : [+1 | -1]
        Propagation mode, +1 (forward) by default 
    input_fft : bool
        Specifies whether input array is fft data or not.
    output_fft : bool
        Specifies whether output array is fft data or not.
    betamax : float
        Defines the betamax parameter if transformation is done in Fourier space.
        
    Returns
    -------
    jones : ndarray
        Output jones array of shape (..., 2,n,m).
    """
    modal = beta is None and phi is None
    
    field = np.asarray(field)
    out = np.empty(field.shape[:-3] + (2,) + field.shape[-2:], field.dtype)
    if input_fft == False and modal:
        field = fft2(field)  
    elif input_fft == True and not modal:
        field = ifft2(field) 
    shape = field.shape[-3:]
    if modal:
        a,fmat = diffraction_alphaf(shape, ks, epsv, epsa, betamax = betamax)
    else:
        a, fmat, = alphaf(beta,phi,epsv,epsa) 
    fvec = field2fvec(field)
    evec = field2fvec(out)
    evec[...] = fvec2E(fvec, fmat = fmat, mode = mode)
    if output_fft == False and modal:
        ifft2(out, out = out)
    elif output_fft == True and not modal:
        fft2(out, out = out)
    return out

def jones2field(jones, ks, beta = None, phi = None, epsv = (1.,1.,1.), epsa = (0.,0.,0.),  mode = +1, input_fft = False, output_fft = False, betamax = BETAMAX):
    """Converts (..., 2,n,m) jones array to (..., 4,n,m) field array
    
    The conversion is done either in reciprocal space (default) or in real space,
    if you provide the beta or phi parameter.
    
    Parameters
    ----------
    jones : ndarray
        Input jones array data of shape (..., 2,n,m).
    ks : array
        A list of k-values (wave numbers).
    beta : float, optional
        If set, it defines the field beta parameter, conversion is then done
        in real space.
    phi : float, optional
        If set, it defines the field phi parameter, conversion is then done
        in real space.
    epsv : array
        The epsilon eigenvalues of the medium
    epsv : array
        The Euler rotation angles.
    mode : [+1 | -1]
        Propagation mode, +1 (forward) by default 
    input_fft : bool
        Specifies whether input array is fft data or not.
    output_fft : bool
        Specifies whether output array is fft data or not.
    betamax : float
        Defines the betamax parameter if transformation is done in Fourier space.
        
    Returns
    -------
    field : ndarray
        Output field array of shape (..., 4,n,m).
    """

    jones= np.asarray(jones)
    modal = beta is None and phi is None
    
    out = np.empty(jones.shape[:-3] + (4,) + jones.shape[-2:], jones.dtype)
    if input_fft == False and modal:
        jones = fft2(jones)   
    elif input_fft == True and not modal:
        jones = ifft2(jones)  
        
    shape = jones.shape[-3:]
    if modal:
        a,fmat = diffraction_alphaf(shape, ks, epsv, epsa, betamax = betamax)
    else:
        a, fmat, = alphaf(beta,phi,epsv,epsa) 
    fvec = field2fvec(out)
    jvec = field2fvec(jones)
    E2fvec(jvec, fmat = fmat, mode = mode, out = fvec)
    if output_fft == False and modal:
        ifft2(out, out = out)
    elif output_fft == True and not modal:
        fft2(out, out = out)
    return out

def _field2modes(field, k0, betamax = BETAMAX):
    f = fft2(field)
    mask = eigenmask(f.shape[-2:], k0, betamax)
    f = f[mask]
    return np.moveaxis(f,-2,-1)

def field2modes(field, k0, betamax = BETAMAX):
    """Converts 2D field array to modes array.
    
    Parameters
    ----------
    field : ndarray or tuple of ndarrays
        Input field array (or tuple of input fields for each wavenumber). The 
        shape of the input arrays must be (...,4,:,:) representing.
    k0 : float or a sequence of floats
        Defines the wavenumber. Fol multi-wavelength data, this must be a
        sequence of wawenumbers
    betamax : float, optional
        The beta cutoff parameter.
        
    Returns
    -------
    mask, modes : ndarray, ndarray or ndarray, tuple of ndarrays
        For a single-wavelength data it returns the mask array specifying
        the mode indices and modes coefficients array. For multi-wavelength data
        the modes is a tuple of ndarrays for each of the wavelengths. Length
        of the mask in this case equals length of the wavenumbers. 
            
    """
    if isinstance(field, tuple):
        out = tuple((field2modes(field[i], k0[i], betamax ) for i in range(len(field))))
        mask = tuple(o[0] for o in out)
        modes = tuple(o[1] for o in out)
        return mask, modes
    
    f = fft2(field)
    k0 = np.asarray(k0)
    mask = eigenmask(f.shape[-2:], k0, betamax)
    if k0.ndim == 0:
        return mask, np.moveaxis(f[...,mask],-2,-1)
    else:
        return mask, tuple((np.moveaxis(f[...,i,:,:,:][...,mask[i]],-2,-1) for i in range(len(k0))))

def ffield2modes(ffield, k0, betamax = BETAMAX):

    k0 = np.asarray(k0)
    mask = eigenmask(ffield.shape[-2:], k0, betamax)
    if k0.ndim == 0:
        return mask, np.moveaxis(ffield[...,mask],-2,-1)
    else:
        return mask, tuple((np.moveaxis(ffield[...,i,:,:,:][...,mask[i]],-2,-1) for i in range(len(k0))))
    

def field2modes1(field, k0, betamax = BETAMAX):
    """Converts field array to modes array.
    
    Parameters
    ----------
    field : ndarray or tuple of ndarrays
        Input field array (or tuple of input fields for each wavenumber). The 
        shape of the input arrays must be (...,4,:) representing.
    k0 : float or a sequence of floats
        Defines the wavenumber. Fol multi-wavelength data, this must be a
        sequence of wawenumbers
    betamax : float, optional
        The beta cutoff parameter.
        
    Returns
    -------
    mask, modes : ndarray, ndarray or ndarray, tuple of ndarrays
        For a single-wavelength data it returns the mask array specifying
        the mode indices and modes coefficients array. For multi-wavelength data
        the modes is a tuple of ndarrays for each of the wavelengths.
            
    """
    if isinstance(field, tuple):
        out = tuple((field2modes1(field[i], k0[i], betamax ) for i in range(len(field))))
        mask = tuple(o[0] for o in out)
        modes = tuple(o[1] for o in out)
        return mask, modes
    
    f = fft(field)
    k0 = np.asarray(k0)
    mask = eigenmask1(f.shape[-1], k0, betamax)
    if k0.ndim == 0:
        return mask, np.moveaxis(f[...,mask],-2,-1)
    else:
        return mask, tuple((np.moveaxis(f[...,i,:,:][...,mask[i]],-2,-1) for i in range(len(k0))))
    
    
def modes2ffield(mask, modes, out = None):
    """Inverse of ffield2modes. Takes the output of ffield2modes and recunstructs
    the field array.
    
    Parameters
    ----------
    mask : ndarray
        Mask array, as returned by field2modes
    modes : ndarray or tuple of ndarrays
        Modes array or a tuple of modes array (in case of multi-wavelength data).
    """
    if isinstance(mask, tuple):
        if out is None:
            out = (None,)* len(mask)
        return tuple((modes2ffield(mask[i], modes[i], out = out[i]) for i in range(len(mask))))

    shape = mask.shape[-2:]
    if mask.ndim == 2:
        shape = modes.shape[:-2] + (4,) + shape
        if out is None:
            out = np.zeros(shape =shape, dtype = CDTYPE )
        else:
            out[...] = 0.
        modes = np.moveaxis(modes,-1,-2)
        out[...,mask] = modes
    else:
        shape = modes[0].shape[:-2] + (len(mask),4,) + shape
        if out is None:
            out = np.zeros(shape =shape, dtype = CDTYPE )
        else:
            out[...] = 0.
        for i,(mode, m) in enumerate(zip(modes,mask)):
            mode = np.moveaxis(mode,-1,-2)
            o = out[...,i,:,:,:]
            o[...,m] = mode
            #print(o[...,m].shape)
            #print(mode.shape)
       
    return out

def modes2field(mask, modes, out = None):
    """Inverse of field2modes. Takes the output of field2modes and recunstructs
    the field array.
    
    Parameters
    ----------
    mask : ndarray
        Mask array, as returned by field2modes
    modes : ndarray or tuple of ndarrays
        Modes array or a tuple of modes array (in case of multi-wavelength data).
    """
    out = modes2ffield(mask, modes, out = out)
    return ifft2(out, out)

def modes2field1(mask, modes):
    """Inverse of field2modes. Takes the output of field2modes and recunstructs
    the field array.
    
    Parameters
    ----------
    mask : ndarray
        Mask array, as returned by field2modes
    modes : ndarray or tuple of ndarrays
        Modes array or a tuple of modes array (in case of multi-wavelength data).
    """
    if isinstance(mask, tuple):
        return tuple((modes2field1(mask[i], modes[i]) for i in range(len(mask))))

    shape = mask.shape[-1:]
    if mask.ndim == 1:
        shape = modes.shape[:-2] + (4,) + shape
        out = np.zeros(shape =shape, dtype = CDTYPE )

        modes = np.moveaxis(modes,-1,-2)
        out[...,mask] = modes
        return ifft(out, overwrite_x = True)
    else:
        shape = modes[0].shape[:-2] + (len(mask),4,) + shape
        out = np.zeros(shape =shape, dtype = CDTYPE )

        for i,(mode, m) in enumerate(zip(modes,mask)):
            mode = np.moveaxis(mode,-1,-2)
            o = out[...,i,:,:]
            o[...,m] = mode
            #print(o[...,m].shape)
            #print(mode.shape)
       
        return ifft(out, overwrite_x = True)

def aperture2rays(diaphragm, betastep = 0.1, norm = True):
    """Takes a 2D image of a diaphragm and converts it to beta, phi, intensity"""
    shape = diaphragm.shape

    ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    xx, yy = np.meshgrid(ax, ay, indexing = "xy")
    phi = np.arctan2(yy,xx)
    beta = np.sqrt(xx**2 + yy**2)*betastep
    mask = diaphragm > 0.
    intensity = diaphragm[mask]
    if norm == True:
        intensity = intensity/ intensity.sum() * len(intensity)
    return np.asarray(beta[mask],FDTYPE), np.asarray(phi[mask],FDTYPE), np.asarray(intensity,FDTYPE)

def illumination_aperture(diameter = 5., smooth = 0.1):
    n = int(round(diameter))
    alpha = min(max(smooth,0),1)
    a = aperture((n,n), diameter/n, alpha)
    return a
  
def illumination_rays(NA, diameter = 5., smooth = 0.1):
    """Returns beta, phi, intensity values for illumination.
    
    This function can be used to define beta,phi,intensiy arrays that can be used to
    construct illumination data with the :func:`illumination_data` function.
    The resulting beta,phi parameters define directions of rays for the input 
    light with a homogeneous angular distribution of rays - input
    light with a given numerical aperture.
    
    Parameters
    ----------
    NA : float
        Approximate numerical aperture of the illumination.
    diameter : int
        Field aperture diaphragm diameter in pixels. Approximate number of rays
        is np.pi*(diameter/2)**2
    smooth : float, optional
        Smoothness of diaphragm edge between 0. and 1.
        
    Returns
    -------
    beta,phi, intensity : ndarrays
        Ray parameters  
    """
    betastep = 2.*NA/(diameter-1)
    a = illumination_aperture(diameter, smooth)
    return aperture2rays(a, betastep = betastep, norm = True)


def illumination_eigenrays(shape, k0, NA = 0.2, scale = (1,1)):
    scale = np.asarray(scale, dtype = int)
    if scale.ndim == 0:
        scale = (scale, scale)
    elif scale.ndim != 1 or len(scale) != 2:
        raise ValueError("Unknown scale parameter!")
    b,p = betaphi(shape,k0)
    bs = b[...,::scale[0],::scale[1]]
    ps = p[...,::scale[0],::scale[1]]
    
    mask = bs <= NA
    
    if bs.ndim == 3:
        bs = tuple((b[m] for b,m in zip(bs,mask)))
        ps = tuple((ps[m] for m in mask))
        intensity = tuple((1./m.sum() for m in mask))
        return bs, ps, intensity
    else:
        return bs[mask], ps[mask], 1./mask.sum()


#def illumination_betaphi(NA, nrays = 13):
#    """Returns beta, phi values for illumination.
#    
#    This function can be used to define beta and phi arrays that can be used to
#    construct illumination data with the :func:`illumination_data` function.
#    The resulting beta,phi parameters define directions of rays for the input 
#    light with a homogeneous angular distribution of rays - input
#    light with a given numerical aperture.
#    
#    Parameters
#    ----------
#    NA : float
#        Approximate numerical aperture of the illumination.
#    nrays : int, optional
#        Approximate number of rays. 
#        
#    Returns
#    -------
#    array, array
#        beta, phi arrays 
#        
#    """
#    radius = (nrays/np.pi)**0.5
#    shape = (1+2*int(radius),1+2*int(radius))
#    ay, ax = [np.arange(-l // 2 + 1., l // 2 + 1.) for l in shape]
#    xx, yy = np.meshgrid(ax, ay,copy = False, indexing = "xy") 
#    phi = np.arctan2(yy,xx)
#    beta = np.sqrt(xx**2 + yy**2)/radius*NA
#    mask = (beta <= NA)
#    return np.asarray(beta[mask],FDTYPE), np.asarray(phi[mask],FDTYPE)

def illumination_waves(shape, k0, beta = 0., phi = 0., window = None, out = None):
    """Builds scalar illumination wave. 
    """
    k0 = np.asarray(k0)
    beta = np.asarray(beta)
    phi = np.asarray(phi)
    if beta.ndim < 1 :#and k0.ndim == 1:
        beta = beta[...,np.newaxis]
    if phi.ndim < 1 :#and k0.ndim == 1:
        phi = phi[...,np.newaxis]        
    if not k0.ndim in (0,1):
        raise ValueError("k0, must be an array with dimesion 1")
    out = planewave(shape, k0, beta, phi, out)
    if window is None:
        return out
    else:
        return np.multiply(out, window, out = out)

def illumination_eigenwaves(shape, k0, beta = 0., phi = 0., out = None):
    """Builds scalar illumination wave. 
    """
    k0 = np.asarray(k0)
    beta = np.asarray(beta)
    phi = np.asarray(phi)
    if beta.ndim < 1 :#and k0.ndim == 1:
        beta = beta[...,np.newaxis]
    if phi.ndim < 1 :#and k0.ndim == 1:
        phi = phi[...,np.newaxis]        
    if not k0.ndim in (0,1):
        raise ValueError("k0, must be an array with dimesion 1")
    out = planewave(shape, k0, beta, phi, out)


def field2betaphi(field_in,ks, multiray = False):
    beta, phi = mean_betaphi(field_in, ks)
    if field_in.ndim > 4 and multiray == True:  #must have at least two polarization states or multi-ray input
        beta = beta.mean(axis = tuple(range(1,field_in.ndim-3))) #average all, but first (multu-ray) axis
        phi = phi.mean(axis = tuple(range(1,field_in.ndim-3)))
    else:
        beta = beta.mean() #mean over all axes - single ray case
        phi = phi.mean()
    return beta, phi

    
def waves2field(waves, k0, beta = 0., phi = 0., n = 1., focus = 0., 
                jones = None, intensity = None, mode = "t", diffraction = True, betamax = BETAMAX):
    """Converts scalar waves to vector field data."""
    beta = np.asarray(beta)
    phi = np.asarray(phi)
    k0 = np.asarray(k0)
    nrays = 1
    if jones is None:
        nrays = 2
    if beta.ndim == 1:
        nrays = len(beta)*nrays
        
    waves = waves/(nrays**0.5)  
    
        
    if jones is None:
        fieldv = np.zeros(beta.shape + (2,) + k0.shape + (4,) + waves.shape[-2:], dtype = CDTYPE)
    else:
        c,s = jones
        fieldv = np.zeros(beta.shape + k0.shape + (4,) + waves.shape[-2:], dtype = CDTYPE)
    
    if beta.ndim == 1: 
        for i,data in enumerate(fieldv):
            if jones is None:
                data[0,...,0,:,:] = waves[i]
                data[0,...,1,:,:] = waves[i]
                
                data[1,...,2,:,:] = waves[i]
                data[1,...,3,:,:] = -waves[i]
            else:
                data[...,0,:,:] = waves[i]*c
                data[...,1,:,:] = waves[i]*c
                data[...,2,:,:] = waves[i]*s
                data[...,3,:,:] = -waves[i]*s
                
    else:
        if jones is None:
            fieldv[0,...,0,:,:] = waves
            fieldv[0,...,1,:,:] = waves
            
            fieldv[1,...,2,:,:] = waves
            fieldv[1,...,3,:,:] = -waves
        else:
            fieldv[...,0,:,:] = waves*c
            fieldv[...,1,:,:] = waves*c
            fieldv[...,2,:,:] = waves*s
            fieldv[...,3,:,:] = -waves*s   
    if diffraction == True:       
        diffracted_field(fieldv,k0, d = -focus, n = n, mode = mode, betamax = betamax, out = fieldv)
    
    #normalize field to these intensities
    if intensity is not None:
        intensity = intensity/nrays
        if jones is None:
            
            #-2 element must be polarization. make it broadcastable to field intensity
            intensity = intensity[...,None,:]
        norm = (intensity/(field2intensity(fieldv).sum((-2,-1))))**0.5
        np.multiply(fieldv, norm[...,None,None,None], fieldv)
        
    return fieldv


def waves2field2(waves, fmat, jones = None, phi = 0,mode = +1):
    """Converts scalar waves to vector field data."""
    if jones is None:
        fvec1 = field4(fmat, jones = jonesvec((1,0),phi), amplitude = waves, mode = mode)
        fvec2 = field4(fmat, jones = jonesvec((0,1),phi), amplitude = waves, mode = mode)
        field1 = fvec2field(fvec1)
        field2 = fvec2field(fvec2)
        
        shape = list(field1.shape)
        shape.insert(-4, 2)
        out = np.empty(shape = shape, dtype = field1.dtype)
        out[...,0,:,:,:,:] = field1
        out[...,1,:,:,:,:] = field2
    else:
        fvec = field4(fmat, jones = jonesvec(jones,phi), amplitude = waves, mode = mode)
        out = fvec2field(fvec).copy()
    return out



def illumination_data(shape, wavelengths, pixelsize = 1., beta = 0., phi = 0., intensity = 1.,
                      n = None, focus = 0., window = None, backdir = False, 
                      jones = None, diffraction = True, eigenmodes = None, betamax = BETAMAX):
    """Constructs forward (or backward) propagating input illumination field data.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the illumination
    wavelengths : array_like
        A list of wavelengths.
    pixelsize : float, optional
        Size of the pixel in nm.
    beta : float or array_like of floats, optional
        Beta parameter(s) of the illumination. (Default 0. - normal incidence) 
    phi : float or array_like of floats, optional
        Azimuthal angle(s) of the illumination. 
    n : float, optional
        Refractive index of the media that this illumination field is assumed to
        be propagating in (default to `n_cover`)
    focus : float, optional
        Focal plane of the field. By default it is set at z=0. 
    window : array or None, optional
        If None, no window function is applied. This window function
        is multiplied with the constructed plane waves to define field diafragm
        of the input light. See :func:`.window.aperture`.
    backdir : bool, optional
        Whether field is bacward propagating, instead of being forward
        propagating (default)
    jones : jones vector or None, optional
        If specified it has to be a valid jones vector that defines polarization
        of the light. If not given (default), the resulting field will have two
        polarization components. See documentation for details and examples.
    diffraction : bool, optional
        Specifies whether field is diffraction limited or not. By default, the 
        field is filtered so that it has only propagating waves. You can disable
        this by specifying diffraction = False.    
    eigenmodes : bool or None
        If set to True (sefault value when `window` = None.), it defines whether 
        to build eigenmodes from beta and phi values. In this case, beta and phi 
        are only approximate values. If set to False (default when `window` != None), 
        true beta and phi values are set. 
    betamax : float, optional
        The betamax parameter of the propagating field.

        
    Returns
    -------
    illumination : field_data
        A field data tuple.       
    """
    
    verbose_level = DTMMConfig.verbose
    if verbose_level > 0:
        print("Building illumination data.") 
        
    n = get_default_config_option("n_cover",n)
    
    eigenmodes = (window is None) if eigenmodes is None else bool(eigenmodes)
    
    wavelengths = np.asarray(wavelengths)
    wavenumbers = 2*np.pi/wavelengths * pixelsize
    if wavenumbers.ndim not in (1,):
        raise ValueError("Wavelengths should be 1D array")
        
    if jones is None:
        intensity = intensity/2.
     
    #intensity = ((np.abs(waves)**2).sum((-2,-1)))* np.asarray(intensity)[...,None]#sum over pixels
    #intensity = intensity * intensity
    mode = -1 if backdir else +1
    _beta = np.asarray(beta, FDTYPE)
    _phi = np.asarray(phi, FDTYPE)
    _intensity = np.asarray(intensity, FDTYPE)

    nrays = len(_beta) if _beta.ndim > 0 else 1
    
    if _beta.ndim == 1:
        beta = _beta[:,None]
    else:
        beta = _beta
    if _phi.ndim == 1:
        phi = _phi[:,None]
    else:
        phi = _phi
        
    if verbose_level > 1:
        print("------------------------------------")
        print(" $ shape: {}".format(shape)) 
        print(" $ num wavelengths: {}".format(len(wavelengths)))  
        print(" $ num rays: {}".format(nrays))  
        print(" $ refractive index: {}".format(n)) 
        print(" $ diffraction limited: {}".format(diffraction))  
        print(" $ using eigenmodes: {}".format(eigenmodes))     
        print(" $ max beta: {}".format(betamax))   
        print(" $ jones: {}".format( jones if jones is not None else "unpolarized")) 
        print("------------------------------------")
        
    waves = illumination_waves(shape, wavenumbers, beta = beta, phi = phi, window = window)
 
    if window is None:
        waves = wave2eigenwave(waves)
     
    if _beta.ndim == 1:
        beta = _beta[:,None,None,None]
    else:
        beta = _beta[...,None,None]
    if _phi.ndim == 1:
        phi = _phi[:,None,None,None] 
    else:
        phi = _phi[...,None,None] 
    intensity = _intensity[...,None,None,None,None]   
    
    epsa = np.asarray((0.,0.,0.),FDTYPE)
    alpha, fmat = alphaf(beta, phi, refind2eps([n]*3), epsa)
    field = waves2field2(waves, fmat, jones = jones, phi = phi, mode = mode)
    intensity1 = field2intensity(field)
    norm = np.ones_like(intensity1)
    norm[:,...] = (intensity/nrays)**0.5    
    if diffraction == True:  
        diffracted_field(field,wavenumbers, d = -focus, n = n, mode = mode, betamax = betamax, out = field)
        intensity2 = field2intensity(field)
        ratio = (intensity1.sum((-2,-1))/intensity2.sum((-2,-1)))**0.5
        norm[...] = norm * ratio[...,None,None]
            
    np.multiply(norm[...,None,:,:], field, field)
    
    return (field, wavelengths, pixelsize)

#def illumination_dataold(shape, wavelengths, pixelsize = 1., beta = 0., phi = 0., intensity = 1.,
#                      n = 1., focus = 0., window = None, backdir = False, 
#                      jones = None, diffraction = True, betamax = BETAMAX):
#    """Constructs forward (or backward) propagating input illumination field data.
#    
#    Parameters
#    ----------
#    shape : (int,int)
#        Shape of the illumination
#    wavelengths : array_like
#        A list of wavelengths.
#    pixelsize : float, optional
#        Size of the pixel in nm.
#    beta : float or array_like of floats, optional
#        Beta parameter(s) of the illumination. (Default 0. - normal incidence) 
#    phi : float or array_like of floats, optional
#        Azimuthal angle(s) of the illumination. 
#    n : float, optional
#        Refractive index of the media that this illumination field is assumed to
#        be propagating in (default 1.)
#    focus : float, optional
#        Focal plane of the field. By default it is set at z=0. 
#    window : array or None, optional
#        If None, no window function is applied. This window function
#        is multiplied with the constructed plane waves to define field diafragm
#        of the input light. See :func:`.window.aperture`.
#    backdir : bool, optional
#        Whether field is bacward propagating, instead of being forward
#        propagating (default)
#    jones : jones vector or None, optional
#        If specified it has to be a valid jones vector that defines polarization
#        of the light. If not given (default), the resulting field will have two
#        polarization components. See documentation for details and examples.
#    diffraction : bool, optional
#        Specifies whether field is diffraction limited or not. By default, the 
#        field is filtered so that it has only propagating waves. You can disable
#        this by specifying diffraction = False.    
#    betamax : float, optional
#        The betamax parameter of the propagating field.
#    """
#    
#    verbose_level = DTMMConfig.verbose
#    if verbose_level > 0:
#        print("Building illumination data.") 
#    wavelengths = np.asarray(wavelengths)
#    wavenumbers = 2*np.pi/wavelengths * pixelsize
#    if wavenumbers.ndim not in (1,):
#        raise ValueError("Wavelengths should be 1D array")
#    waves = illumination_waves(shape, wavenumbers, beta = beta, phi = phi, window = window)
#    intensity = ((np.abs(waves)**2).sum((-2,-1)))* np.asarray(intensity)[...,None]#sum over pixels
#    mode = "r" if backdir else "t"
#    field = waves2field(waves, wavenumbers, intensity = intensity, beta = beta, phi = phi, n = n,
#                        focus = focus, jones = jones, mode = mode, betamax = betamax)
#    return (field, wavelengths, pixelsize)


@nb.njit([(NCDTYPE[:,:,:],NFDTYPE[:,:])], cache = NUMBA_CACHE)
def _field2intensity(field, out):
    for j in range(field.shape[1]):
        for k in range(field.shape[2]):
            tmp1 = (field[0,j,k].real * field[1,j,k].real + field[0,j,k].imag * field[1,j,k].imag)
            tmp2 = (field[2,j,k].real * field[3,j,k].real + field[2,j,k].imag * field[3,j,k].imag)
            out[j,k] = tmp1-tmp2 

@nb.guvectorize([(NCDTYPE[:,:,:],NFDTYPE[:,:])],"(k,n,m)->(n,m)", target = NUMBA_TARGET, cache = NUMBA_CACHE)
def field2intensity(field, out):
    """field2intensity(field)
    
Converts field array of shape [...,4,height,width] to intensity array
of shape [...,height,width]. For each pixel element, a normal
component of the Poynting vector is computed.
    
Parameters
----------
field : array_like
    Input field array
cmf : array_like
    Color matching function
    
Returns
-------
spec : ndarray
    Computed intensity array"""  
    assert len(field) == 4
    _field2intensity(field, out)

@nb.njit([(NCDTYPE[:,:,:,:],NFDTYPE[:,:,:])], parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE)
def _field2specter(field, out):     
    for j in prange(field.shape[2]):
        for i in range(field.shape[0]):
            for k in range(field.shape[3]):
                tmp1 = (field[i,0,j,k].real * field[i,1,j,k].real + field[i,0,j,k].imag * field[i,1,j,k].imag)
                tmp2 = (field[i,2,j,k].real * field[i,3,j,k].real + field[i,2,j,k].imag * field[i,3,j,k].imag)
                out[j,k,i] = tmp1-tmp2 

@nb.njit([(NCDTYPE[:,:,:,:,:],NFDTYPE[:,:,:])], parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE)
def _field2spectersum(field, out):
    for n in prange(field.shape[0]):
        for j in range(field.shape[3]):
            for i in range(field.shape[1]):
                for k in range(field.shape[4]):
                    tmp1 = (field[n,i,0,j,k].real * field[n,i,1,j,k].real + field[n,i,0,j,k].imag * field[n,i,1,j,k].imag)
                    tmp2 = (field[n,i,2,j,k].real * field[n,i,3,j,k].real + field[n,i,2,j,k].imag * field[n,i,3,j,k].imag)
                    if n == 0:
                        out[j,k,i] = tmp1 - tmp2
                    else:
                        out[j,k,i] += (tmp1 -tmp2)

@nb.guvectorize([(NCDTYPE[:,:,:,:],NFDTYPE[:,:,:])],"(w,k,n,m)->(n,m,w)", target = "cpu", cache = NUMBA_CACHE)
def field2specter(field, out):
    """field2specter(field)
    
Converts field array of shape [...,nwavelengths,4,height,width] to specter array
of shape [...,height,width,nwavelengths]. For each pixel element, a normal
componentof Poynting vector is computed
    
Parameters
----------
field : array_like
    Input field array
cmf : array_like
    Color matching function
    
Returns
-------
spec : ndarray
    Computed specter array""" 
    _field2specter(field, out)

@nb.guvectorize([(NCDTYPE[:,:,:,:,:],NFDTYPE[:,:,:])],"(l,w,k,n,m)->(n,m,w)", target = "cpu", cache = NUMBA_CACHE)
def field2spectersum(field, out):
    _field2spectersum(field, out)  
    
    
@nb.guvectorize([(NCDTYPE[:,:,:],NFDTYPE[:,:],NFDTYPE[:,:],NFDTYPE[:],NFDTYPE[:])], "(k,n,m),(n,m),(n,m)->(),()", target = NUMBA_TARGET, cache = NUMBA_CACHE)
def _fft_betaphi(f, betax, betay, beta, phi):

    _betax = 0.
    _betay = 0.
    _ssum = 0.
    for i in range(f.shape[0]):
        for j in range(f.shape[1]):
            for k in range(f.shape[2]):
                s = f[i,j,k].real**2 + f[i,j,k].imag**2
                _betax += s*betax[j,k]
                _betay += s*betay[j,k]
                _ssum += s
    #save results to output
    beta[0] = (_betax**2+_betay**2)**0.5 / _ssum
    phi[0] = np.arctan2(_betay,_betax)

def mean_betaphi(field,k0):
    """Calculates mean beta and phi of a given field."""
    b = blackman(field.shape[-2:])
    f = fft2(field*b) #filter it with blackman..
    betax, betay = betaxy(field.shape[-2:], k0)
    beta, phi = _fft_betaphi(f,betax,betay)
    return beta, phi

def validate_field_data(data):
    """Validates field data.
    
    This function inspects validity of the field data, and makes proper data
    conversions to match the field data format. In case data is not valid and 
    it cannot be converted to a valid data it raises an exception (ValueError). 
    
    Parameters
    ----------
    data : tuple of field data
        A valid field data tuple.
    
    Returns
    -------
    data : tuple
        Validated field data tuple. 
    """
    field, wavelengths, pixelsize = data
    field = np.asarray(field, dtype = CDTYPE)
    wavelengths = np.asarray(wavelengths, dtype = FDTYPE)
    pixelsize = float(pixelsize)
    if field.ndim < 4:
        raise ValueError("Invald field dimensions")
    if field.shape[-4] != len(wavelengths) or wavelengths.ndim != 1:
        raise ValueError("Incompatible wavelengths shape.")

    return field, wavelengths, pixelsize


MAGIC = b"dtmf" #legth 4 magic number for file ID
VERSION = b"\x00"

def save_field(file, field_data):
    """Saves field data to a binary file in ``.dtmf`` format.
    
    Parameters
    ----------
    file : file, str, or pathlib.Path
        File or filename to which the data is saved.  If file is a file-object,
        then the filename is unchanged.  If file is a string, a ``.dtmf``
        extension will be appended to the file name if it does not already
        have one.
    field_data: (field,wavelengths,pixelsize)
        A valid field data tuple
    
    """    
    own_fid = False
    field, wavelengths,pixelsize = validate_field_data(field_data)
    try:
        if isinstance(file, str):
            if not file.endswith('.dtmf'):
                file = file + '.dtmf'
            f = open(file, "wb")
            own_fid = True
        else:
            f = file
        f.write(MAGIC)
        f.write(VERSION)
        np.save(f,field)
        np.save(f,wavelengths)
        np.save(f,pixelsize)
    finally:
        if own_fid == True:
            f.close()


def load_field(file):
    """Load field data from file
    
    Parameters
    ----------
    file : file, str
        The file or filenam to read.
    """
    own_fid = False
    try:
        if isinstance(file, str):
            f = open(file, "rb")
            own_fid = True
        else:
            f = file
        magic = f.read(len(MAGIC))
        if magic == MAGIC:
            if f.read(1) != VERSION:
                raise OSError("This file was created with a more recent version of dtmm. Please upgrade your dtmm package!")
            field = np.load(f)
            wavelengths = np.load(f)
            pixelsize = float(np.load(f))
            return field, wavelengths, pixelsize
        else:
            raise OSError("Failed to interpret file {}".format(file))
    finally:
        if own_fid == True:
            f.close()
field2poynting = field2intensity
    
__all__ = ["illumination_rays","load_field", "save_field", "validate_field_data","field2specter","field2intensity", "illumination_data"]